# -*- coding: utf-8 -*-

from StringIO import StringIO
from colander import null
from pyramid.httpexceptions import HTTPMovedPermanently

from mock import MagicMock
import pytest

from kotti.testing import DummyRequest
from kotti.testing import asset
from kotti.views.file import inline_view
from kotti.views.file import attachment_view
from kotti.views.file import UploadedFileResponse
from kotti.filedepot import StoredFileResponse


class TestFileViews:
    def _create_file(self, config):
        from kotti.resources import File
        self.file = File(asset('logo.png').read(), u"myfüle.png", u"image/png")

    def _test_common_headers(self, headers):
        for name in ('Content-Disposition', 'Content-Length', 'Content-Type'):
            assert isinstance(headers[name], str)
        assert headers["Content-Length"] == "55715"
        assert headers["Content-Type"] == "image/png"

    @pytest.mark.parametrize("params",
                             [(inline_view, 'inline'),
                              (attachment_view, 'attachment')])
    def test_file_views(self, params, config, filedepot, dummy_request,
                        depot_tween):
        from kotti.filedepot import TweenFactory

        view, disposition = params
        self._create_file(config)

        res = view(self.file, dummy_request)

        self._test_common_headers(res.headers)

        assert res.content_disposition.startswith(
            '{0}; filename=my'.format(disposition))

        assert res.app_iter.file.read() == asset('logo.png').read()
        res.app_iter.file.seek(0)
        assert res.body == asset('logo.png').read()


class TestFileEditForm:
    def make_one(self):
        from kotti.views.edit.content import FileEditForm
        from kotti.resources import File

        return FileEditForm(File(), DummyRequest())

    def test_edit_with_file(self, db_session, filedepot):
        view = self.make_one()
        view.edit(
            title=u'A title', description=u'A description',
            tags=[u"A tag"],
            file=dict(
                fp=StringIO('filecontents'),
                filename=u'myfile.png',
                mimetype=u'image/png',
                size=10,
                uid="randomabc",
                ),
            )
        assert view.context.title == u'A title'
        assert view.context.description == u'A description'
        assert view.context.data.file.read() == 'filecontents'
        assert view.context.filename == u'myfile.png'
        assert view.context.mimetype == u'image/png'
        assert view.context.size == len('filecontents')
        assert view.context.tags == [u"A tag"]

    def test_edit_without_file(self, filedepot):
        view = self.make_one()
        view.context.data = 'filecontents'
        view.context.filename = u'myfile.png'
        view.context.mimetype = u'image/png'
        view.context.size = 777
        view.edit(title=u'A title',
                  description=u'A description',
                  tags=[],
                  file=null)
        assert view.context.title == u'A title'
        assert view.context.description == u'A description'
        assert view.context.data.file.read() == 'filecontents'
        assert view.context.filename == u'myfile.png'
        assert view.context.mimetype == u'image/png'
        assert view.context.size == 777


class TestFileAddForm:
    def make_one(self):
        from kotti.views.edit.content import FileAddForm
        return FileAddForm(MagicMock(), DummyRequest())

    def test_add(self, config, filedepot):
        view = self.make_one()
        file = view.add(
            title=u'A title',
            description=u'A description',
            tags=[],
            file=dict(
                fp=StringIO('filecontents'),
                filename=u'myfile.png',
                mimetype=u'image/png',
                size=None,
                uid="randomabc",
                ),
            )

        assert file.title == u'A title'
        assert file.description == u'A description'
        assert file.tags == []
        assert file.data.file.read() == 'filecontents'
        assert file.filename == u'myfile.png'
        assert file.mimetype == u'image/png'
        assert file.size == len('filecontents')


class TestFileUploadTempStore:
    def make_one(self):
        from kotti.views.form import FileUploadTempStore
        return FileUploadTempStore(DummyRequest())

    def test_keys(self):
        tmpstore = self.make_one()
        tmpstore.session['important'] = 3
        tmpstore.session['_secret'] = 4
        assert tmpstore.keys() == ['important']

    def test_delitem(self):
        tmpstore = self.make_one()
        tmpstore.session['important'] = 3
        del tmpstore['important']
        assert 'important' not in tmpstore.session


class TestDepotStore:

    from kotti.resources import File, Image

    @pytest.mark.parametrize("factory", [File, Image])
    def test_create(self, factory, filedepot, image_asset, app):
        data = image_asset.read()
        f = factory(data)
        if factory.__class__.__name__ == 'File':
            assert len(f.data['files']) == 1
        elif factory.__class__.__name__ == 'Image':
            assert len(f.data['files']) == 4
        assert f.data.file.read() == data

    @pytest.mark.parametrize("factory", [File, Image])
    def test_edit_content(self, factory, filedepot, image_asset, image_asset2,
                          app, db_session):
        data = image_asset.read()
        f = factory(data)
        assert f.data.file.read() == data
        db_session.flush()
        data2 = image_asset2.read()
        f.data = data2
        db_session.flush()

        assert f.data.file.read() == data2

    @pytest.mark.parametrize("factory", [File, Image])
    def test_session_rollback(self, factory, db_session, filedepot, image_asset,
                              app):
        from depot.manager import DepotManager

        f = factory(data=image_asset.read(), name=u'content', title=u'content')
        id = f.data['file_id']

        db_session.add(f)
        db_session.flush()
        assert id in DepotManager.get()._storage.keys()

        db_session.rollback()
        assert id not in DepotManager.get()._storage.keys()
        assert DepotManager.get().delete.called

    @pytest.mark.parametrize("factory", [File, Image])
    def test_delete(self, factory, db_session, root, filedepot, image_asset,
                    app):
        from depot.manager import DepotManager

        f = factory(data=image_asset, name=u'content', title=u'content')
        id = f.data['file_id']
        root[str(id)] = f
        db_session.flush()

        assert id in DepotManager.get()._storage.keys()

        del root[str(id)]
        import transaction
        transaction.commit()

        assert DepotManager.get().delete.called
        assert id not in DepotManager.get()._storage.keys()


class TestUploadedFileResponse:
    def _create_file(self,
                     data="file contents",
                     filename=u"myfüle.png",
                     mimetype=u"image/png"):
        from kotti.resources import File
        return File(data, filename, mimetype)

    def test_as_body(self, filedepot, image_asset):
        data = image_asset.read()
        f = self._create_file(data)
        resp = UploadedFileResponse(f.data, DummyRequest())
        assert resp.body == data

    def test_as_app_iter(self, filedepot, image_asset):
        from pyramid.response import FileIter
        data = image_asset.read()
        f = self._create_file(data)
        resp = UploadedFileResponse(f.data, DummyRequest())
        assert isinstance(resp.app_iter, FileIter)
        assert ''.join(resp.app_iter) == data

    def test_unknown_filename(self, filedepot, image_asset2):
        f = self._create_file(b'foo', u"file.bar", None)
        resp = UploadedFileResponse(f.data, DummyRequest())
        assert resp.headers['Content-Type'] == 'application/octet-stream'

    def test_guess_content_type(self, filedepot, image_asset):
        f = self._create_file(image_asset.read(), u"file.png", None)
        resp = UploadedFileResponse(f.data, DummyRequest())
        assert resp.headers['Content-Type'] == 'image/png'

    def test_caching(self, filedepot, monkeypatch):
        import datetime
        import webob.response

        f = self._create_file()
        d = datetime.datetime(2012, 12, 31, 13)

        class mockdatetime:
            @staticmethod
            def utcnow():
                return d

        monkeypatch.setattr(webob.response, 'datetime', mockdatetime)

        resp = UploadedFileResponse(f.data, DummyRequest(), cache_max_age=10)

        # this is set by filedepot fixture
        assert resp.headers['Last-Modified'] == 'Sun, 30 Dec 2012 00:00:00 GMT'
        assert resp.headers['Expires'] == 'Mon, 31 Dec 2012 13:00:10 GMT'

    def test_redirect(self, filedepot):

        class PublicFile(object):
            public_url = 'http://example.com'

        class PublicData(object):
            file = PublicFile()

        with pytest.raises(HTTPMovedPermanently) as e:
            UploadedFileResponse(PublicData(), DummyRequest())

        response = e.value
        assert response.headers['Location'] == 'http://example.com'


class TestStoredFileResponse:
    def _create_file(self,
                     data="file contents",
                     filename=u"myfüle.png",
                     mimetype=u"image/png"):
        from kotti.resources import File
        return File(data, filename, mimetype)

    def test_as_body(self, filedepot, image_asset):
        data = image_asset.read()
        f = self._create_file(data)
        resp = StoredFileResponse(f.data.file, DummyRequest())
        assert resp.body == data

    def test_as_app_iter(self, filedepot, image_asset):
        from pyramid.response import FileIter
        data = image_asset.read()
        f = self._create_file(data)
        resp = StoredFileResponse(f.data.file, DummyRequest())
        assert isinstance(resp.app_iter, FileIter)
        assert ''.join(resp.app_iter) == data

    def test_unknown_filename(self, filedepot, image_asset2):
        f = self._create_file(b'foo', u"file.bar", None)
        resp = StoredFileResponse(f.data.file, DummyRequest())
        assert resp.headers['Content-Type'] == 'application/octet-stream'

    def test_guess_content_type(self, filedepot, image_asset):
        f = self._create_file(image_asset.read(), u"file.png", None)
        resp = StoredFileResponse(f.data.file, DummyRequest())
        assert resp.headers['Content-Type'] == 'image/png'

    def test_caching(self, filedepot, monkeypatch):
        import datetime
        import webob.response

        f = self._create_file()
        d = datetime.datetime(2012, 12, 31, 13)

        class mockdatetime:
            @staticmethod
            def utcnow():
                return d

        monkeypatch.setattr(webob.response, 'datetime', mockdatetime)

        resp = StoredFileResponse(f.data.file, DummyRequest(), cache_max_age=10)

        # this is set by filedepot fixture
        assert resp.headers['Last-Modified'] == 'Sun, 30 Dec 2012 00:00:00 GMT'
        assert resp.headers['Expires'] == 'Mon, 31 Dec 2012 13:00:10 GMT'

    def test_redirect(self, filedepot):
        class PublicFile(object):
            public_url = 'http://example.com'

        with pytest.raises(HTTPMovedPermanently) as e:
            StoredFileResponse(PublicFile(), DummyRequest())

        response = e.value
        assert response.headers['Location'] == 'http://example.com'



# -*- coding: utf-8 -*-
"""
Form related base views from which you can inherit.

Inheritance Diagram
-------------------

.. inheritance-diagram:: kotti.views.form
"""

from StringIO import StringIO
from UserDict import DictMixin

import colander
import deform.widget
from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPFound
from pyramid_deform import CSRFSchema
from pyramid_deform import FormView

from kotti import get_settings
from kotti.fanstatic import tagit
from kotti.resources import Tag
# noinspection PyProtectedMember
from kotti.util import _
from kotti.util import title_to_name
from kotti.util import translate


def get_appstruct(context, schema):
    appstruct = {}
    for field in schema.children:
        if hasattr(context, field.name):
            val = getattr(context, field.name)
            if val is None:
                val = colander.null
            appstruct[field.name] = val
    return appstruct


class ObjectType(colander.SchemaType):
    """ A type leaving the value untouched.
    """

    @staticmethod
    def serialize(node, value):
        return value

    @staticmethod
    def deserialize(node, value):
        return value


@colander.deferred
def deferred_tag_it_widget(node, kw):
    tagit.need()
    all_tags = Tag.query.all()
    available_tags = [tag.title.encode('utf-8') for tag in all_tags]
    widget = CommaSeparatedListWidget(template='tag_it',
                                      available_tags=available_tags)
    return widget


class Form(deform.Form):
    """
    A deform Form that allows 'appstruct' to be set on the instance.
    """

    def render(self, appstruct=None, readonly=False):
        if appstruct is None:
            appstruct = getattr(self, 'appstruct', colander.null)
        return super(Form, self).render(appstruct, readonly=readonly)


class BaseFormView(FormView):
    """
    A basic view for forms with save and cancel buttons.
    """

    form_class = Form
    buttons = (
        deform.Button('save', _(u'Save')),
        deform.Button('cancel', _(u'Cancel')))
    success_message = _(u"Your changes have been saved.")
    success_url = None
    schema_factory = None
    use_csrf_token = True
    add_template_vars = ()

    def __init__(self, context, request, **kwargs):
        super(BaseFormView, self).__init__(request)
        self.context = context
        self.__dict__.update(kwargs)

    def __call__(self):
        if self.schema_factory is not None:
            self.schema = self.schema_factory()
        if self.use_csrf_token and 'csrf_token' not in self.schema:
            self.schema.children.append(CSRFSchema()['csrf_token'])
        result = super(BaseFormView, self).__call__()
        if isinstance(result, dict):
            result.update(self.more_template_vars())
        return result

    def cancel_success(self, appstruct):
        location = self.request.resource_url(self.context)
        return HTTPFound(location=location)
    cancel_failure = cancel_success

    def more_template_vars(self):
        result = {}
        for name in self.add_template_vars:
            result[name] = getattr(self, name)
        return result


class EditFormView(BaseFormView):
    """
    A base form for content editing purposes.

    Set `self.schema_factory` to the context's schema.  Values of
    fields in this schema will be set as attributes on the context.
    An example::

        import colander
        from deform.widget import RichTextWidget

        from kotti.edit.content import ContentSchema
        from kotti.edit.content import EditFormView

        class DocumentSchema(ContentSchema):
            body = colander.SchemaNode(
                colander.String(),
                title=u'Body',
                widget=RichTextWidget(),
                missing=u'',
                )

        class DocumentEditForm(EditFormView):
            schema_factory = DocumentSchema
    """

    add_template_vars = ('first_heading',)

    def before(self, form):
        form.appstruct = get_appstruct(self.context, self.schema)

    def save_success(self, appstruct):
        appstruct.pop('csrf_token', None)
        self.edit(**appstruct)
        self.request.session.flash(self.success_message, 'success')
        location = self.success_url or self.request.resource_url(self.context)
        return HTTPFound(location=location)

    def edit(self, **appstruct):
        for key, value in appstruct.items():
            setattr(self.context, key, value)

    @reify
    def first_heading(self):
        return _(u'Edit ${title}',
                 mapping=dict(title=self.context.title)
                 )


class AddFormView(BaseFormView):
    """
    A base form for content adding purposes.

    Set `self.schema_factory` as with EditFormView.  Also set
    `item_type` to your model class.  An example::

        class DocumentAddForm(AddFormView):
            schema_factory = DocumentSchema
            add = Document
            item_type = u'Document'
    """

    success_message = _(u"Item was added.")
    item_type = None
    add_template_vars = ('first_heading',)

    def save_success(self, appstruct):
        appstruct.pop('csrf_token', None)
        name = self.find_name(appstruct)
        new_item = self.context[name] = self.add(**appstruct)
        self.request.session.flash(self.success_message, 'success')
        location = self.success_url or self.request.resource_url(new_item)
        return HTTPFound(location=location)

    def find_name(self, appstruct):
        name = appstruct.get('name')
        if name is None:
            name = title_to_name(
                appstruct['title'], blacklist=self.context.keys())
        return name

    @reify
    def first_heading(self):
        context_title = getattr(self.request.context, 'title', None)
        type_title = self.item_type or self.add.type_info.title
        if context_title:
            return _(u'Add ${type} to ${title}.',
                     mapping=dict(type=translate(type_title),
                                  title=context_title))
        else:
            return _(u'Add ${type}.', mapping=dict(type=translate(type_title)))


class CommaSeparatedListWidget(deform.widget.Widget):

    def __init__(self, template=None, **kw):
        super(CommaSeparatedListWidget, self).__init__(**kw)
        self.template = template

    def serialize(self, field, cstruct, readonly=False):
        if cstruct in (colander.null, None):
            cstruct = []
        return field.renderer(self.template, field=field, cstruct=cstruct)

    # noinspection PyMethodOverriding
    @staticmethod
    def deserialize(field, pstruct):
        if pstruct is colander.null:
            return colander.null
        return [item.strip() for item in pstruct.split(',') if item]


class FileUploadTempStore(DictMixin):
    """
    A temporary storage for file file uploads

    File uploads are stored in the session so that you don't need
    to upload your file again if validation of another schema node
    fails.
    """

    def __init__(self, request):
        self.session = request.session

    def keys(self):
        return [k for k in self.session.keys() if not k.startswith('_')]

    def __setitem__(self, name, value):
        value = value.copy()
        fp = value.pop('fp')
        value['file_contents'] = fp.read()
        fp.seek(0)
        self.session[name] = value

    def __getitem__(self, name):
        value = self.session[name].copy()
        value['fp'] = StringIO(value.pop('file_contents'))
        return value

    def __delitem__(self, name):
        del self.session[name]

    @staticmethod
    def preview_url(name):
        return None


def validate_file_size_limit(node, value):
    """
    File size limit validator.

    You can configure the maximum size by setting the kotti.max_file_size
    option to the maximum number of bytes that you want to allow.
    """

    value['fp'].seek(0, 2)
    size = value['fp'].tell()
    value['fp'].seek(0)
    max_size = get_settings()['kotti.max_file_size']
    if size > int(max_size) * 1024 * 1024:
        msg = _('Maximum file size: ${size}MB', mapping={'size': max_size})
        raise colander.Invalid(node, msg)



# -*- coding: utf-8 -*-

from pyramid.view import view_config

from kotti.interfaces import IFile


@view_config(name='view', context=IFile, permission='view',
             renderer='kotti:templates/view/file.pt')
def view(context, request):
    return {}


@view_config(name='inline-view', context=IFile, permission='view')
def inline_view(context, request):
    return request.uploaded_file_response(context.data)


@view_config(name='attachment-view', context=IFile, permission='view')
def attachment_view(context, request):
    return request.uploaded_file_response(context.data, 'attachment')


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.scan(__name__)


# DEPRECATED

# noinspection PyPep8
from zope.deprecation.deprecation import deprecated
# noinspection PyPep8
from kotti.filedepot import StoredFileResponse


class UploadedFileResponse(StoredFileResponse):
    def __init__(self, data, request, disposition='attachment',
                 cache_max_age=None, content_type=None,
                 content_encoding=None):
        super(UploadedFileResponse, self).__init__(
            data.file, request, disposition=disposition,
            cache_max_age=cache_max_age, content_type=content_type,
            content_encoding=content_encoding)

deprecated('UploadedFileResponse',
           'UploadedFileResponse is deprecated and will be removed in '
           'Kotti 2.0.0.  Use "request.uploaded_file_response(context.data)" '
           'instead.')



# -*- coding: utf-8 -*-
"""
:summary: Default view selctor views
"""

import warnings

from pyramid.compat import map_
from pyramid.httpexceptions import HTTPFound
from pyramid.interfaces import IView
from pyramid.interfaces import IViewClassifier
from pyramid.threadlocal import get_current_registry
from pyramid.view import view_config
from pyramid.view import view_defaults
from zope.interface import providedBy

# noinspection PyProtectedMember
from kotti.util import _


@view_defaults(permission='edit')
class DefaultViewSelection(object):

    def __init__(self, context, request):

        self.context = context
        self.request = request

    def _get_view(self, view_name):  # pragma: no cover
        """This code is copied from pyramid.view.
           We trust it and don't test.

           Returns True if a view with name view_name is registered for context.
        """
        provides = [IViewClassifier] + map_(
            providedBy,
            (self.request, self.context)
        )

        try:
            reg = self.request.registry
        except AttributeError:
            reg = get_current_registry()

        return reg.adapters.lookup(provides, IView, name=view_name)

    def _is_valid_view(self, view_name):
        """Return True if a view with name view_name is registered for context.
        """

        return self._get_view(view_name) is not None

    def _is_valid_default_view(self, view_name):
        """Return True if a view with name view_name is a valid choice as
        default view for context.
        """

        return self._is_valid_view(view_name) and view_name in [
            v[0] for v in self.context.type_info.selectable_default_views]

    @view_config(name='default-view-selector',
                 renderer='kotti:templates/default-view-selector.pt')
    def default_view_selector(self):
        """
        :summary: Submenu for selection of the node's default view.
        :result: Dictionary with a selectable_default_views list.
        :rtype: dict
        """

        sviews = []

        for v in self.context.type_info.selectable_default_views:
            name, title = v
            if self._is_valid_view(name):
                sviews.append({
                    "name": name,
                    "title": title,
                    "is_current": name == self.context.default_view,
                })
            else:
                warnings.warn(
                    u"No view called '{0}' is registered for {1!r}.".format(
                        name, self.context))

        return {
            "selectable_default_views": [
                {
                    "name": "default",
                    "title": _("Default view"),
                    "is_current": self.context.default_view is None,
                }
            ] + sviews,
        }

    @view_config(name='set-default-view')
    def set_default_view(self):
        """
        :summary: Set the node's default view and redirect to it.
        :result: Redirect to the context URL.
        :rtype: pyramid.httpexceptions.HTTPFound
        """

        if 'view_name' in self.request.GET:

            view_name = self.request.GET['view_name']

            if view_name == "default":
                self.context.default_view = None
                self.request.session.flash(
                    _("Default view has been reset to default."),
                    'success'
                )
            else:
                if self._is_valid_default_view(view_name):
                    self.context.default_view = view_name
                    self.request.session.flash(
                        _("Default view has been set."),
                        'success'
                    )
                else:
                    self.request.session.flash(
                        _("Default view could not be set."),
                        'error'
                    )

        return HTTPFound(
            location=self.request.resource_url(self.context)
        )


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.scan('.default_views')



# -*- coding: utf-8 -*-

"""
Login / logout and forbidden views and forms.
"""

from datetime import datetime

import colander
from deform import Button
from deform import Form
from deform import ValidationFailure
from deform.widget import CheckedPasswordWidget
from deform.widget import HiddenWidget
from formencode.validators import Email
from pyramid.encode import urlencode
from pyramid.httpexceptions import HTTPForbidden
from pyramid.httpexceptions import HTTPFound
from pyramid.security import forget
from pyramid.security import remember
from pyramid.url import resource_url
from pyramid.view import view_config
from pyramid.settings import asbool

from kotti import get_settings
from kotti.message import email_set_password
from kotti.message import validate_token
from kotti.security import get_principals
# noinspection PyProtectedMember
from kotti.util import _
from kotti.views.util import template_api
from kotti.views.users import deferred_email_validator
from kotti.views.users import name_pattern_validator
from kotti.views.users import name_new_validator
from kotti.views.users import UserAddFormView
from kotti.events import ObjectEvent
from kotti.events import notify


def _find_user(login):
    principals = get_principals()
    principal = principals.get(login)
    if principal is not None:
        return principal
    else:
        # noinspection PyBroadException
        try:
            Email().to_python(login)
        except Exception:
            pass
        else:
            for p in principals.search(email=login):
                return p


class UserSelfRegistered(ObjectEvent):
    """ This event is emitted just after user self registered. Intended use
        is to allow addons to do some preparation for such user - create custom
        contents, nodes etc.
        Event handler object parameter is a Principal object
    """


class RegisterSchema(colander.Schema):
    title = colander.SchemaNode(
        colander.String(),
        title=_(u'Full name'))
    name = colander.SchemaNode(
        colander.String(),
        title=_(u'Username'),
        validator=colander.All(name_pattern_validator, name_new_validator)
    )
    email = colander.SchemaNode(
        colander.String(),
        title=_(u'Email'),
        validator=deferred_email_validator,
    )


@view_config(name='register', renderer='kotti:templates/edit/simpleform.pt',
             if_setting_has_value=('kotti.register', True))
def register(context, request):
    schema = RegisterSchema().bind(request=request)
    form = Form(schema, buttons=(Button('register', _(u'Register')),))
    rendered_form = None

    if 'register' in request.POST:
        try:
            appstruct = form.validate(request.POST.items())
        except ValidationFailure as e:
            request.session.flash(_(u"There was an error."), 'error')
            rendered_form = e.render()
        else:
            settings = get_settings()

            appstruct['groups'] = u''
            appstruct['roles'] = u''

            register_groups = settings['kotti.register.group']
            if register_groups:
                appstruct['groups'] = [register_groups]

            register_roles = settings['kotti.register.role']
            if register_roles:
                appstruct['roles'] = {'role:' + register_roles}

            appstruct['send_email'] = True
            form = UserAddFormView(context, request)
            form.add_user_success(appstruct)
            success_msg = _(
                'Congratulations! You are successfully registered. '
                'You should be receiving an email with a link to set your '
                'password. Doing so will activate your account.'
                )
            request.session.flash(success_msg, 'success')
            name = appstruct['name']
            notify(UserSelfRegistered(get_principals()[name], request))
            return HTTPFound(location=request.application_url)

    if rendered_form is None:
        rendered_form = form.render(request.params)

    api = template_api(
        context, request,
        page_title=_(u"Register - ${title}",
                     mapping=dict(title=context.title)),
    )

    return {
        'api': api,
        'form': rendered_form,
        }


def login_success_callback(request, user, came_from):
    """ Default implementation of ``kotti.login_success_callback``.  You can
    implement a custom function with the same signature and point the
    ``kotti.login_success_callback`` setting to it.

    :param request: Current request
    :type request: :class:`kotti.request.Request`

    :param user: Principal, who just logged in successfully.
    :type user: :class:`kotti.security.Princial`

    :param came_from: URL the user came from
    :type came_from: str

    :result: Any Pyramid response object, by default a redirect to
             ``came_from`` or the context where login was called.
    :rtype: :class:`pyramid.httpexceptions.HTTPFound`
    """

    headers = remember(request, user.name)
    request.session.flash(
        _(u"Welcome, ${user}!",
          mapping=dict(user=user.title or user.name)), 'success')
    user.last_login_date = datetime.now()
    return HTTPFound(location=came_from, headers=headers)


def reset_password_callback(request, user):
    """ Default implementation of ``kotti.reset_password_callback``.  You can
    implement a custom function with the same signature and point the
    ``kotti.reset_password_callback`` setting to it.

    :param request: Current request
    :type request: :class:`kotti.request.Request`

    :param user: Principal, who's password was requested to be reset.
    :type user: :class:`kotti.security.Princial`

    :result: Any Pyramid response object, by default a redirect to to the same
             URL from where the password reset was called.
    :rtype: :class:`pyramid.httpexceptions.HTTPFound`
    """

    email_set_password(
        user, request,
        template_name='kotti:templates/email-reset-password.pt')
    request.session.flash(_(
        u"You should be receiving an email with a link to reset your "
        u"password. Doing so will activate your account."), 'success')

    return HTTPFound(location=request.url)


@view_config(name='login', renderer='kotti:templates/login.pt')
def login(context, request):
    """
    Login view.  Renders either the login or password forgot form templates or
    handles their form submission and redirects to came_from on success.

    :result: Either a redirect response or a dictionary passed to the template
             for rendering
    :rtype: pyramid.httpexceptions.HTTPFound or dict
    """

    principals = get_principals()

    came_from = request.params.get(
        'came_from', request.resource_url(context))
    login, password = u'', u''

    if 'submit' in request.POST:
        login = request.params['login'].lower()
        password = request.params['password']
        user = _find_user(login)

        if (user is not None and user.active and
                principals.validate_password(password, user.password)):
            return get_settings()['kotti.login_success_callback'][0](
                request, user, came_from)
        request.session.flash(_(u"Login failed."), 'error')

    if 'reset-password' in request.POST:
        login = request.params['login']
        user = _find_user(login)
        if user is not None and user.active:
            return get_settings()['kotti.reset_password_callback'][0](
                request, user)
        else:
            request.session.flash(
                _(u"That username or email is not known by this system."),
                'error')

    return {
        'url': request.application_url + '/@@login',
        'came_from': came_from,
        'login': login,
        'password': password,
        'register': asbool(get_settings()['kotti.register']),
        }


@view_config(name='logout')
def logout(context, request):
    """
    Logout view.  Always redirects the user to where he came from.

    :result: Redirect to came_from
    :rtype: pyramid.httpexceptions.HTTPFound
    """

    headers = forget(request)
    request.session.flash(_(u"You have been logged out."), 'info')
    location = request.params.get('came_from', request.application_url)
    return HTTPFound(location=location, headers=headers)


class SetPasswordSchema(colander.MappingSchema):
    """
    Schema for the set password form
    """

    #: colander.String
    password = colander.SchemaNode(
        colander.String(),
        title=_(u'Password'),
        validator=colander.Length(min=5),
        widget=CheckedPasswordWidget(),
        )
    #: colander.String
    token = colander.SchemaNode(
        colander.String(),
        widget=HiddenWidget(),
        )
    #: colander.String
    email = colander.SchemaNode(
        colander.String(),
        title=_(u'Email'),
        widget=HiddenWidget(),
        )
    #: colander.String
    continue_to = colander.SchemaNode(
        colander.String(),
        widget=HiddenWidget(),
        missing=colander.null,
        )


@view_config(name='set-password', renderer='kotti:templates/edit/simpleform.pt')
def set_password(context, request,
                 success_msg=_(u"You have reset your password.")):
    """ Set password view.  Displays the set password form and handles its form
    submission.

    :param context: Current context
    :type context: :class:`kotti.resources.Content`

    :param request: Current request
    :type request: :class:`kotti.request.Request`

    :param success_msg: Message to display on successful submission handling
    :type success_msg: str or TranslationString

    :result: Redirect response or dictionary passed to the template for
             rendering.
    :rtype: pyramid.httpexceptions.HTTPFound or dict
    """

    form = Form(SetPasswordSchema(),
                buttons=(Button('submit', _(u'Set password')),))
    rendered_form = None

    if 'submit' in request.POST:
        try:
            appstruct = form.validate(request.POST.items())
        except ValidationFailure as e:
            rendered_form = e.render()
        else:
            token = appstruct['token']
            email = appstruct['email']
            user = _find_user(email)
            if (user is not None and
                    validate_token(user, token) and
                    token == user.confirm_token and
                    user.active):
                password = appstruct['password']
                user.password = get_principals().hash_password(password)
                user.confirm_token = None
                headers = remember(request, user.name)
                user.last_login_date = datetime.now()

                location = (appstruct['continue_to'] or
                            resource_url(context, request))
                request.session.flash(success_msg, 'success')
                return HTTPFound(location=location, headers=headers)
            else:
                request.session.flash(
                    _(u"Your password reset token may have expired."), 'error')

    if rendered_form is None:
        rendered_form = form.render(request.params)

    api = template_api(
        context, request,
        page_title=_(u"Reset your password - ${title}.",
                     mapping=dict(title=context.title)),
        )

    return {
        'api': api,
        'form': rendered_form,
        }


@view_config(context=HTTPForbidden, accept='text/html',)
def forbidden_redirect(context, request):
    """ Forbidden redirect view.  Redirects to the login form for anonymous
    users or to the forbidden view for authenticated users.

    :result: Redirect to one of the above.
    :rtype: pyramid.httpexceptions.HTTPFound
    """
    if request.authenticated_userid:
        location = request.application_url + '/@@forbidden'
    else:
        location = request.application_url + '/@@login?' + urlencode(
            {'came_from': request.url})
    return HTTPFound(location=location)


@view_config(context=HTTPForbidden)
def forbidden_view(request):
    """
    Forbidden view.  Raises 403 for requests not originating from a web browser
    like device.

    :result: 403
    :rtype: pyramid.httpexceptions.HTTPForbidden
    """

    return request.exception


@view_config(name='forbidden', renderer='kotti:templates/forbidden.pt')
def forbidden_view_html(request):
    """
    Forbidden view for browsers.

    :result: empty dictionary passed to the template for rendering
    :rtype: dict
    """

    return {}


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.scan(__name__)



# -*- coding: utf-8 -*-
#
# Kotti documentation build configuration file, created by
# sphinx-quickstart on Fri Mar 18 23:14:47 2011.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import pkg_resources
from datetime import date

# -- General configuration -----------------------------------------------------

extensions = [
    'repoze.sphinx.autointerface',
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.graphviz',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    ]

copyright = u'2012-{0}, Kotti developers'.format(date.today().year)
exclude_patterns = ['_build', '_themes']
master_doc = 'index'
project = u'Kotti'
pygments_style = 'sphinx'
source_suffix = '.rst'
templates_path = ['_templates']
today_fmt = '%Y-%m-%d'
# The full version, including alpha/beta/rc tags.
version = pkg_resources.get_distribution("Kotti").version
release = version

# add_function_parentheses = True
# add_module_names = True
# default_role = None
# language = None
# modindex_common_prefix = []
# show_authors = False
# source_encoding = 'utf-8-sig'

# -- Options for HTML output ---------------------------------------------------

# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get('READTHEDOCS') == 'True'

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# A shorter title for the navigation bar.  Default is the same as html_title.
html_last_updated_fmt = '%Y-%m-%d'
html_static_path = ['_static']
htmlhelp_basename = 'Kottidoc'
# html_additional_pages = {}
# html_domain_indices = True
# html_favicon = None
# html_file_suffix = None
# html_logo = None
# html_short_title = None
# html_show_copyright = True
# html_show_sourcelink = True
# html_show_sphinx = True
# html_sidebars = {}
# html_split_index = False
# html_title = None
# html_use_index = True
# html_use_opensearch = ''
# html_use_smartypants = True


# -- Options for LaTeX output --------------------------------------------------

# latex_appendices = []
# latex_documents = [
#     ('index', 'Kotti.tex', u'Kotti Documentation', u'Daniel Nouri', 'manual'),
# ]
# latex_domain_indices = True
# latex_font_size = '10pt'
# latex_logo = None
# latex_paper_size = 'a4'
# latex_preamble = ''
# latex_show_pagerefs = False
# latex_show_urls = False
# latex_use_parts = False


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# man_pages = [
#     ('index', 'kotti', u'Kotti Documentation', [u'Daniel Nouri'], 1)
# ]

# -- Options for inheritance diagrams ------------------------------------------
inheritance_graph_attrs = dict(
    rankdir='TB',
    nodesep=0.1,
    ratio='auto',
    size=11.0
)
inheritance_node_attrs = dict(
    height=0.7,
    margin='"0.06, 0.03"',
    fontname='"Vera Sans, DejaVu Sans, Liberation Sans, Arial, Helvetica, sans"'
)
graphviz_output_format = 'svg'

# -- Options for Intersphinx ---------------------------------------------------
intersphinx_mapping = {
    'bleach': ('https://bleach.readthedocs.io/en/latest/', None),
    'colander': ('https://colander.readthedocs.io/en/latest/', None),
    'deform': ('https://deform.readthedocs.io/en/latest/', None),
    'depot': ('https://depot.readthedocs.io/en/latest/', None),
    'fanstatic': ('http://www.fanstatic.org/en/latest/', None),
    'pyramid': ('https://pyramid.readthedocs.io/en/latest/', None),
    'sqlalchemy': ('https://sqlalchemy.readthedocs.io/en/latest/', None),
}



# -*- coding: utf-8 -*-
from pytest import raises


def test_needed_group():

    from js.deform import deform_js
    from kotti.fanstatic import contents_view_js
    from kotti.fanstatic import NeededGroup

    def NeededGroupFactory(resources):
        return NeededGroup(resources)

    with raises(ValueError):
        NeededGroupFactory("foo")
    with raises(ValueError):
        NeededGroupFactory(["foo", "bar"])

    needed = NeededGroup([deform_js, ])

    assert needed.resources == [deform_js, ]

    needed.add(contents_view_js)

    assert needed.resources == [deform_js, contents_view_js]

    def needed_group_adder(resource):
        needed.add(resource)

    with raises(ValueError):
        needed_group_adder(42)

    needed = NeededGroup()
    assert needed.resources == []



# -*- coding: utf-8 -*-
"""Enlarged name title

Revision ID: 37a05f6246af
Revises: 413fa5fcc581
Create Date: 2015-05-15 17:12:07.631190

"""

# revision identifiers, used by Alembic.
revision = '37a05f6246af'
down_revision = '413fa5fcc581'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('nodes',
                    'title',
                    type_=sa.Unicode(250),
                    )
    op.alter_column('nodes',
                    'name',
                    type_=sa.Unicode(250),
                    )
    op.alter_column('nodes',
                    'path',
                    type_=sa.Unicode(2000),
                    )


def downgrade():
    op.alter_column('nodes',
                    'title',
                    type_=sa.Unicode(100),
                    )
    op.alter_column('nodes',
                    'name',
                    type_=sa.Unicode(50),
                    )
    op.alter_column('nodes',
                    'path',
                    type_=sa.Unicode(1000),
                    )



# -*- coding: utf-8 -*-
from kotti.testing import DummyRequest
from mock import patch
from mock import call


class TestRegister:

    def test_register_form(self, root):
        from kotti.views.login import register

        request = DummyRequest()
        res = register(root, request)
        assert(res['form'][:5] == '<form')

    def test_register_submit_empty(self, root):
        from kotti.views.login import register

        request = DummyRequest()
        request.POST['register'] = u'register'
        res = register(root, request)
        assert 'There was a problem with your submission' in res['form']

    def test_register_submit(self, root):
        from kotti.views.login import register
        from pyramid.httpexceptions import HTTPFound

        request = DummyRequest()
        request.POST['title'] = u'Test User'
        request.POST['name'] = u'test'
        request.POST['email'] = u'test@example.com'
        request.POST['register'] = u'register',

        with patch('kotti.views.login.UserAddFormView') as form:
            with patch('kotti.views.login.get_principals'):
                res = register(root, request)
                form.assert_has_calls([call().add_user_success({
                    'name': u'test',
                    'roles': u'',
                    'title': u'Test User',
                    'send_email': True,
                    'groups': u'',
                    'email': u'test@example.com'})]
                )
        assert(isinstance(res, HTTPFound))

    def test_register_event(self, root):
        from kotti.views.login import register
        request = DummyRequest()
        request.POST['title'] = u'Test User'
        request.POST['name'] = u'test'
        request.POST['email'] = u'test@example.com'
        request.POST['register'] = u'register',

        with patch('kotti.views.login.UserAddFormView'):
            with patch('kotti.views.login.get_principals'):
                with patch('kotti.views.login.notify') as notify:
                    register(root, request)
        assert(notify.call_count == 1)

    def test_register_submit_groups_and_roles(self, root):
        from kotti.views.login import register
        from pyramid.httpexceptions import HTTPFound

        request = DummyRequest()
        request.POST['title'] = u'Test User'
        request.POST['name'] = u'test'
        request.POST['email'] = u'test@example.com'
        request.POST['register'] = u'register',

        with patch('kotti.views.login.UserAddFormView') as form:
            with patch('kotti.views.login.get_principals'):
                with patch('kotti.views.login.get_settings') as get_settings:
                    get_settings.return_value = {
                        'kotti.register.group': 'mygroup',
                        'kotti.register.role': 'myrole',
                        }

                    res = register(root, request)

        form.assert_has_calls([
            call().add_user_success({
                'name': u'test',
                'roles': {u'role:myrole'},
                'title': u'Test User',
                'send_email': True,
                'groups': [u'mygroup'],
                'email': u'test@example.com',
                })])
        assert(isinstance(res, HTTPFound))


class TestNotRegister:
    def test_it(self, app):
        res = app.post('/register', status=404)
        assert res.status == '404 Not Found'



# -*- coding: utf-8 -*-
"""Add ForeignKey indices

Revision ID: 4a3de0d0804a
Revises: 37a05f6246af
Create Date: 2015-08-31 12:37:26.493958

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '4a3de0d0804a'
down_revision = '37a05f6246af'


def upgrade():
    op.create_index(
        'ix_nodes_parent_id', 'nodes', ['parent_id', ])
    op.create_index(
        'ix_local_groups_node_id', 'local_groups', ['node_id', ])
    op.create_index(
        'ix_local_groups_principal_name', 'local_groups', ['principal_name', ])
    op.create_index(
        'ix_tags_to_contents_tag_id', 'tags_to_contents', ['tag_id', ])
    op.create_index(
        'ix_tags_to_contents_content_id', 'tags_to_contents', ['content_id', ])


def downgrade():
    op.drop_index('ix_nodes_parent_id', 'nodes')
    op.drop_index('ix_local_groups_node_id', 'local_groups')
    op.drop_index('ix_local_groups_principal_name', 'local_groups')



# -*- coding: utf-8 -*-

import pyramid.request
from pyramid.decorator import reify
from pyramid.interfaces import IRequest
from zope.interface import implementer

from kotti.security import get_user


@implementer(IRequest)
class Request(pyramid.request.Request):
    """ Kotti subclasses :class:`pyramid.request.Request` to make additional
    attributes / methods available on request objects and override Pyramid's
    :meth:`pyramid.request.Request.has_permission`.  The latter is needed to
    support Kotti's concept of local roles not just for users but also for
    groups (:func:`kotti.security.list_groups_callback`).
    """

    @reify
    def user(self):
        """ Add the authenticated user to the request object.

        :result: the currently authenticated user
        :rtype: :class:`kotti.security.Principal` or whatever is returned by
                the custom principals database defined in the
                ``kotti.principals_factory`` setting
        """
        return get_user(self)

    def has_permission(self, permission, context=None):
        """ Check if the current request has the given permission on the
        current or explicitly passed context.  This is different from
        :meth:`pyramid.request.Request.has_permission`` in that a context other
        than the one bound to the request can be passed.  This allows to
        consider local roles for the check.

        :param permission: name of the permission to check
        :type permission: str

        :param context: context for which the permission is checked.
                        Defaults to the context on which the request invoked.
        :type context: :class:`kotti.resources.Node`

        :result: True if has_permission, False else
        :rtype: bool
        """

        from kotti.security import authz_context

        with authz_context(context, self):
            return super(Request, self).has_permission(permission, context)



# -*- coding: utf-8 -*-
import logging
import mimetypes
import uuid
from datetime import datetime

import rfc6266
from depot.fields.sqlalchemy import _SQLAMutationTracker
from depot.io.interfaces import FileStorage
from depot.manager import DepotManager
from pyramid import tweens
from pyramid.httpexceptions import HTTPMovedPermanently
from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import _BLOCK_SIZE
from pyramid.response import FileIter
from pyramid.response import Response
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import event
from sqlalchemy import Integer
from sqlalchemy import LargeBinary
from sqlalchemy import String
from sqlalchemy import Unicode
from sqlalchemy.orm import deferred
from unidecode import unidecode

from kotti import Base, get_settings
from kotti import DBSession
from kotti.util import camel_case_to_name
from kotti.util import command
from kotti.util import extract_from_settings
from kotti.util import _to_fieldstorage

_marker = object()


class DBStoredFile(Base):
    """ :class:`depot.io.interfaces.StoredFile` implementation that stores
    file data in SQL database.

    Can be used together with :class:`kotti.filedepot.DBFileStorage` to
    implement blobs storage in the database.
    """

    __tablename__ = "blobs"

    #: Primary key column in the DB
    #: (:class:`sqlalchemy.types.Integer`)
    id = Column(Integer(), primary_key=True)
    #: Unique file id given to this blob
    #: (:class:`sqlalchemy.types.String`)
    file_id = Column(String(36), index=True)
    #: The original filename it had when it was uploaded.
    #: (:class:`sqlalchemy.types.String`)
    filename = Column(Unicode(100))
    #: MIME type of the blob
    #: (:class:`sqlalchemy.types.String`)
    content_type = Column(String(100))
    #: Size of the blob in bytes
    #: (:class:`sqlalchemy.types.Integer`)
    content_length = Column(Integer())
    #: Date / time the blob was created or last modified
    #: (:class:`sqlalchemy.types.DateTime`)
    last_modified = Column(DateTime())
    #: The binary data itself
    #: (:class:`sqlalchemy.types.LargeBinary`)
    data = deferred(Column('data', LargeBinary()))

    _cursor = 0
    _data = _marker

    public_url = None

    def __init__(self, file_id, filename=None, content_type=None,
                 last_modified=None, content_length=None, **kwds):
        self.file_id = file_id
        self.filename = filename
        self.content_type = content_type
        self.last_modified = last_modified or datetime.now()
        self.content_length = content_length

        for k, v in kwds.items():
            setattr(self, k, v)

    def read(self, n=-1):
        """Reads ``n`` bytes from the file.

        If ``n`` is not specified or is ``-1`` the whole
        file content is read in memory and returned
        """
        if self._data is _marker:
            file_id = DBSession.merge(self).file_id
            self._data = DBSession.query(DBStoredFile.data).\
                filter_by(file_id=file_id).scalar()

        if n == -1:
            result = self._data[self._cursor:]
        else:
            result = self._data[self._cursor:self._cursor + n]

        self._cursor += len(result)

        return result

    @staticmethod
    def close(*args, **kwargs):
        """Implement :meth:`StoredFile.close`.
        :class:`DBStoredFile` never closes.
        """
        return

    @staticmethod
    def closed():
        """Implement :meth:`StoredFile.closed`.
        """
        return False

    @staticmethod
    def writable():
        """Implement :meth:`StoredFile.writable`.
        """
        return False

    @staticmethod
    def seekable():
        """Implement :meth:`StoredFile.seekable`.
        """
        return True

    def seek(self, offset, whence=0):
        """ Change stream position.

        Change the stream position to the given byte offset. The offset is
        interpreted relative to the position indicated by whence.

        :param offset: Position for the cursor
        :type offset: int

        :param whence: * 0 -- start of stream (the default);
                              offset should be zero or positive
                       * 1 -- current stream position; offset may be negative
                       * 2 -- end of stream; offset is usually negative
        :type whence: int
        """
        if whence == 0:
            self._cursor = offset
        elif whence in (1, 2):
            self._cursor = self._cursor + offset
        else:
            raise ValueError('whence must be 0, 1 or 2')

    def tell(self):
        """ Returns current position of file cursor

        :result: Current file cursor position.
        :rtype: int
        """
        return self._cursor

    @property
    def name(self):
        """Implement :meth:`StoredFile.name`.

        :result: the filename of the saved file
        :rtype: string
        """
        return self.filename

    @classmethod
    def __declare_last__(cls):
        """ Executed by SQLAlchemy as part of mapper configuration

        When the data changes, we want to reset the cursor position of target
        instance, to allow proper streaming of data.
        """
        event.listen(DBStoredFile.data, 'set', handle_change_data)


def handle_change_data(target, value, oldvalue, initiator):
    target._cursor = 0
    target._data = _marker


def set_metadata(event):
    """Set DBStoredFile metadata based on data

    :param event: event that trigerred this handler.
    :type event: :class:`ObjectInsert` or :class:`ObjectUpdate`
    """
    obj = event.object
    obj.content_length = obj.data and len(obj.data) or 0
    obj.last_modified = datetime.now()


class DBFileStorage(FileStorage):
    """Implementation of :class:`depot.io.interfaces.FileStorage`,

    Uses `kotti.filedepot.DBStoredFile` to store blob data in an SQL database.
    """

    # noinspection PyMethodOverriding
    @staticmethod
    def get(file_id):
        """Returns the file given by the file_id

        :param file_id: the unique id associated to the file
        :type file_id: string
        :result: a :class:`kotti.filedepot.DBStoredFile` instance
        :rtype: :class:`kotti.filedepot.DBStoredFile`
        """

        f = DBSession.query(DBStoredFile).filter_by(file_id=file_id).first()
        if f is None:
            raise IOError
        return f

    def create(self, content, filename=None, content_type=None):
        """Saves a new file and returns the file id

        :param content: can either be ``bytes``, another ``file object``
                        or a :class:`cgi.FieldStorage`. When ``filename`` and
                        ``content_type``  parameters are not provided they are
                        deducted from the content itself.

        :param filename: filename for this file
        :type filename: string

        :param content_type: Mimetype of this file
        :type content_type: string

        :return: the unique ``file_id`` associated to this file
        :rtype: string
        """
        new_file_id = str(uuid.uuid1())
        content, filename, content_type = self.fileinfo(
            content, filename, content_type)
        if hasattr(content, 'read'):
            content = content.read()

        fstore = DBStoredFile(data=content,
                              file_id=new_file_id,
                              filename=filename,
                              content_type=content_type,
                              )
        DBSession.add(fstore)
        return new_file_id

    def replace(self, file_or_id, content, filename=None, content_type=None):
        """Replaces an existing file, an ``IOError`` is raised if the file
        didn't already exist.

        Given a :class:`StoredFile` or its ID it will replace the current
        content with the provided ``content`` value. If ``filename`` and
        ``content_type`` are provided or can be deducted by the ``content``
        itself they will also replace the previous values, otherwise
        the current values are kept.

        :param file_or_id: can be either ``DBStoredFile`` or a ``file_id``

        :param content: can either be ``bytes``, another ``file object``
                        or a :class:`cgi.FieldStorage`. When ``filename`` and
                        ``content_type`` parameters are not provided they are
                        deducted from the content itself.

        :param filename: filename for this file
        :type filename: string

        :param content_type: Mimetype of this file
        :type content_type: string
        """

        file_id = self._get_file_id(file_or_id)

        content, filename, content_type = self.fileinfo(
            content, filename, content_type)

        fstore = self.get(file_id)

        if filename is not None:
            fstore.filename = filename
        if content_type is not None:
            fstore.content_type = content_type

        if hasattr(content, 'read'):
            content = content.read()

        fstore.data = content

    def delete(self, file_or_id):
        """Deletes a file. If the file didn't exist it will just do nothing.

        :param file_or_id: can be either ``DBStoredFile`` or a ``file_id``
        """

        file_id = self._get_file_id(file_or_id)

        DBSession.query(DBStoredFile).filter_by(file_id=file_id).delete()

    def exists(self, file_or_id):
        """Returns if a file or its ID still exist.

        :return: Returns if a file or its ID still exist.
        :rtype: bool
        """

        file_id = self._get_file_id(file_or_id)

        return bool(
            DBSession.query(DBStoredFile).filter_by(file_id=file_id).count())

    @staticmethod
    def list(*args):
        raise NotImplementedError("list() method is unimplemented.")

    @staticmethod
    def _get_file_id(file_or_id):
        if hasattr(file_or_id, 'file_id'):
            return file_or_id.file_id
        return file_or_id


def migrate_storage(from_storage, to_storage):

    log = logging.getLogger(__name__)

    old_default = DepotManager._default_depot
    DepotManager._default_depot = to_storage

    for klass, props in _SQLAMutationTracker.mapped_entities.items():
        log.info("Migrating %r", klass)

        mapper = klass._sa_class_manager.mapper

        # use type column to avoid polymorphism issues, getting the same
        # Node item multiple times.
        type_ = camel_case_to_name(klass.__name__)
        for instance in DBSession.query(klass).filter_by(type=type_):
            for prop in props:
                uf = getattr(instance, prop)
                if not uf:
                    continue
                pk = mapper.primary_key_from_instance(instance)
                log.info("Migrating %s for %r with pk %r", prop, klass, pk)

                filename = uf['filename']
                content_type = uf['content_type']
                data = _to_fieldstorage(fp=uf.file,
                                        filename=filename,
                                        mimetype=content_type,
                                        size=uf.file.content_length)
                setattr(instance, prop, data)

    DepotManager._default_depot = old_default


def migrate_storages_command():  # pragma: no cover
    __doc__ = """ Migrate blobs between two configured filedepot storages

    Usage:
      kotti-migrate-storage <config_uri> --from-storage <name> --to-storage <name>

    Options:
      -h --help                 Show this screen.
      --from-storage <name>   The storage name that has blob data to migrate
      --to-storage <name>     The storage name where we want to put the blobs
    """
    return command(
        lambda args: migrate_storage(
            from_storage=args['--from-storage'],
            to_storage=args['--to-storage'],
        ),
        __doc__,
    )


class StoredFileResponse(Response):
    """ A Response object that can be used to serve an UploadedFile instance.

    Code adapted from :class:`pyramid.response.FileResponse`.
    """

    def __init__(self, f, request, disposition='attachment',
                 cache_max_age=604800, content_type=None,
                 content_encoding=None):
        """
        :param f: the ``UploadedFile`` file field value.
        :type f: :class:`depot.io.interfaces.StoredFile`

        :param request: Current request.
        :type request: :class:`pyramid.request.Request`

        :param disposition:
        :type disposition:

        :param cache_max_age: The number of seconds that should be used to HTTP
                              cache this response.

        :param content_type: The content_type of the response.

        :param content_encoding: The content_encoding of the response.
                                 It's generally safe to leave this set to
                                 ``None`` if you're serving a binary file.
                                 This argument will be ignored if you also
                                 leave ``content-type`` as ``None``.
        """

        if f.public_url:
            raise HTTPMovedPermanently(f.public_url)

        content_encoding, content_type = self._get_type_and_encoding(
            content_encoding, content_type, f)

        super(StoredFileResponse, self).__init__(
            conditional_response=True,
            content_type=content_type,
            content_encoding=content_encoding)

        app_iter = None
        if request is not None and \
                not get_settings()['kotti.depot_replace_wsgi_file_wrapper']:
            environ = request.environ
            if 'wsgi.file_wrapper' in environ:
                app_iter = environ['wsgi.file_wrapper'](f, _BLOCK_SIZE)
        if app_iter is None:
            app_iter = FileIter(f)
        self.app_iter = app_iter

        # assignment of content_length must come after assignment of app_iter
        self.content_length = f.content_length
        self.last_modified = f.last_modified

        if cache_max_age is not None:
            self.cache_expires = cache_max_age
            self.cache_control.public = True

        self.etag = self.generate_etag(f)
        self.content_disposition = rfc6266.build_header(
            f.filename, disposition=disposition,
            filename_compat=unidecode(f.filename))

    @staticmethod
    def _get_type_and_encoding(content_encoding, content_type, f):
        content_type = content_type or getattr(f, 'content_type', None)
        if content_type is None:
            content_type, content_encoding = \
                mimetypes.guess_type(f.filename, strict=False)
        if content_type is None:
            content_type = 'application/octet-stream'
        # str-ifying content_type is a workaround for a bug in Python 2.7.7
        # on Windows where mimetypes.guess_type returns unicode for the
        # content_type.
        content_type = str(content_type)
        return content_encoding, content_type

    @staticmethod
    def generate_etag(f):
        return '"{0}-{1}"'.format(f.last_modified, f.content_length)


def uploaded_file_response(self, uploaded_file, disposition='inline'):
    return StoredFileResponse(uploaded_file.file, self,
                              disposition=disposition)


def uploaded_file_url(self, uploaded_file, disposition='inline'):
    if disposition == 'attachment':
        suffix = '/download'
    else:
        suffix = ''
    url = '{0}/{1}/{2}{3}'.format(
        self.application_url,
        get_settings()['kotti.depot_mountpoint'][1:],
        uploaded_file.path,
        suffix)
    return url


class TweenFactory(object):
    """Factory for a Pyramid tween in charge of serving Depot files.

    This is the Pyramid tween version of
    :class:`depot.middleware.DepotMiddleware`.  It does exactly the same as
    Depot's WSGI middleware, but operates on a :class:`pyramid.request.Request`
    object instead of the WSGI environment.
    """

    def __init__(self, handler, registry):
        """
        :param handler: Downstream tween or main Pyramid request handler (Kotti)
        :type handler: function

        :param registry: Application registry
        :type registry: :class:`pyramid.registry.Registry`
        """

        self.mountpoint = registry.settings['kotti.depot_mountpoint']
        self.handler = handler
        self.registry = registry

        DepotManager.set_middleware(self)

    def url_for(self, path):
        return u'/'.join((self.mountpoint, path))

    def __call__(self, request):
        """
        :param request: Current request
        :type request: :class:`kotti.request.Request`

        :return: Respone object
        :rtype: :class:`pyramid.response.Response`
        """

        # Only handle GET and HEAD requests for the mountpoint.
        # All other requests are passed to downstream handlers.
        if request.method not in ('GET', 'HEAD') \
                or not request.path.startswith(self.mountpoint):
            response = self.handler(request)
            return response

        # paths match this pattern
        # /<mountpoint>/<depot name>/<file id>[/download]
        path = request.path.split('/')
        if len(path) and not path[0]:
            path = path[1:]

        if len(path) < 3:
            response = HTTPNotFound()
            return response

        __, depot, fileid = path[:3]
        depot = DepotManager.get(depot)
        if not depot:
            response = HTTPNotFound()
            return response

        try:
            f = depot.get(fileid)
        except (IOError, ValueError):
            response = HTTPNotFound()
            return response

        # if the file has a public_url, it's stored somewhere else (e.g. S3)
        public_url = f.public_url
        if public_url is not None:
            response = HTTPMovedPermanently(public_url)
            return response

        # file is not directly accessible for user agents, serve it ourselves
        if path[-1] == 'download':
            disposition = 'attachment'
        else:
            disposition = 'inline'
        response = StoredFileResponse(f, request, disposition=disposition)
        return response


def adjust_for_engine(conn, branch):
    # adjust for engine type

    if conn.engine.dialect.name == 'mysql':  # pragma: no cover
        from sqlalchemy.dialects.mysql.base import LONGBLOB
        DBStoredFile.__table__.c.data.type = LONGBLOB()

    # sqlite's Unicode columns return a buffer which can't be encoded by
    # a json encoder. We have to convert to a unicode string so that the value
    # can be saved corectly by
    # :class:`depot.fields.sqlalchemy.upload.UploadedFile`

    def patched_processed_result_value(self, value, dialect):
        if not value:
            return None
        return self._upload_type.decode(unicode(value))

    if conn.engine.dialect.name == 'sqlite':  # pragma: no cover
        from depot.fields.sqlalchemy import UploadedFileField
        UploadedFileField.process_result_value = patched_processed_result_value


def extract_depot_settings(prefix="kotti.depot.", settings=None):
    """ Merges items from a dictionary that have keys that start with `prefix`
    to a list of dictionaries.

    :param prefix: A dotted string representing the prefix for the common values
    :type prefix: string

    :param settings: A dictionary with settings. Result is extracted from this
    :type settings: dict

      >>> settings = {
      ...     'kotti.depot_mountpoint': '/depot',
      ...     'kotti.depot.0.backend': 'kotti.filedepot.DBFileStorage',
      ...     'kotti.depot.0.file_storage': 'var/files',
      ...     'kotti.depot.0.name': 'local',
      ...     'kotti.depot.1.backend': 'depot.io.gridfs.GridStorage',
      ...     'kotti.depot.1.name': 'mongodb',
      ...     'kotti.depot.1.uri': 'localhost://',
      ... }
      >>> res = extract_depot_settings('kotti.depot.', settings)
      >>> print(sorted(res[0].items()))
      [('backend', 'kotti.filedepot.DBFileStorage'), ('file_storage', 'var/files'), ('name', 'local')]
      >>> print(sorted(res[1].items()))
      [('backend', 'depot.io.gridfs.GridStorage'), ('name', 'mongodb'), ('uri', 'localhost://')]
    """

    extracted = {}
    for k, v in extract_from_settings(prefix, settings).items():
        index, conf = k.split('.', 1)
        index = int(index)
        extracted.setdefault(index, {})
        extracted[index][conf] = v

    result = []
    for k in sorted(extracted.keys()):
        result.append(extracted[k])

    return result


def configure_filedepot(settings):

    config = extract_depot_settings('kotti.depot.', settings)
    for conf in config:
        name = conf.pop('name')
        if name not in DepotManager._depots:
            DepotManager.configure(name, conf, prefix='')


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.add_tween('kotti.filedepot.TweenFactory',
                     over=tweens.MAIN,
                     under=tweens.INGRESS)
    config.add_request_method(uploaded_file_response,
                              name='uploaded_file_response')
    config.add_request_method(uploaded_file_url, name='uploaded_file_url')

    from kotti.events import objectevent_listeners
    from kotti.events import ObjectInsert
    from kotti.events import ObjectUpdate

    from sqlalchemy.event import listen
    from sqlalchemy.engine import Engine

    listen(Engine, 'engine_connect', adjust_for_engine)

    configure_filedepot(config.get_settings())

    # Update file metadata on change of blob data
    objectevent_listeners[
        (ObjectInsert, DBStoredFile)].append(set_metadata)
    objectevent_listeners[
        (ObjectUpdate, DBStoredFile)].append(set_metadata)

    # depot's _SQLAMutationTracker._session_committed is executed on
    # after_commit, that's too late for DBFileStorage to interact with the
    # session
    event.listen(DBSession,
                 'before_commit',
                 _SQLAMutationTracker._session_committed)



# -*- coding: utf-8 -*-

"""

Fixture dependencies
--------------------

.. graphviz::

   digraph kotti_fixtures {
      "allwarnings";
      "app" -> "webtest";
      "config" -> "db_session";
      "config" -> "depot_tween";
      "config" -> "dummy_request";
      "config" -> "events";
      "config" -> "workflow";
      "connection" -> "content";
      "connection" -> "db_session";
      "content" -> "db_session";
      "custom_settings" -> "connection";
      "custom_settings" -> "unresolved_settings";
      "db_session" -> "app";
      "db_session" -> "browser";
      "db_session" -> "filedepot";
      "db_session" -> "root";
      "depot_tween" -> "webtest";
      "dummy_mailer" -> "app";
      "dummy_mailer";
      "dummy_request" -> "depot_tween";
      "events" -> "app";
      "depot_tween" -> "filedepot";
      "depot_tween" -> "mock_filedepot";
      "mock_filedepot";
      "depot_tween" -> "no_filedepots";
      "settings" -> "config";
      "settings" -> "content";
      "setup_app" -> "app";
      "setup_app" -> "browser";
      "unresolved_settings" -> "settings";
      "unresolved_settings" -> "setup_app";
      "workflow" -> "app";
   }

"""

# public pytest fixtures

import warnings
from datetime import datetime

from pytest import fixture
from mock import MagicMock

from kotti import testing


@fixture
def image_asset():
    """ Return an image file """

    return testing.asset('sendeschluss.jpg')


@fixture
def image_asset2():
    """ Return another image file """

    return testing.asset('logo.png')


@fixture
def allwarnings(request):
    save_filters = warnings.filters[:]
    warnings.filters[:] = []

    def restore():
        warnings.filters[:] = save_filters

    request.addfinalizer(restore)


@fixture(scope='session')
def custom_settings():
    """ This is a dummy fixture meant to be overriden in add on package's
    ``conftest.py``.  It can be used to inject arbitrary settings for third
    party test suites.  The default settings dictionary will be updated
    with the dictionary returned by this fixture.

    This is also a good place to import your add on's ``resources`` module to
    have the corresponding tables created during ``create_all()`` in
    :func:`kotti.tests.content`.

    :result: settings
    :rtype: dict
    """

    return {}


@fixture(scope='session')
def unresolved_settings(custom_settings):
    from kotti import conf_defaults
    from kotti.testing import testing_db_url
    settings = conf_defaults.copy()
    settings['kotti.secret'] = 'secret'
    settings['kotti.secret2'] = 'secret2'
    settings['kotti.populators'] = 'kotti.testing._populator'
    settings['sqlalchemy.url'] = testing_db_url()
    settings.update(custom_settings)
    return settings


@fixture(scope='session')
def settings(unresolved_settings):
    from kotti import _resolve_dotted
    return _resolve_dotted(unresolved_settings)


@fixture
def config(request, settings):
    """ returns a Pyramid `Configurator` object initialized
        with Kotti's default (test) settings.
    """
    from pyramid import testing
    from kotti import security
    config = testing.setUp(settings=settings)
    config.include('pyramid_chameleon')
    config.add_default_renderers()
    request.addfinalizer(security.reset)
    request.addfinalizer(testing.tearDown)
    return config


@fixture(scope='session')
def connection(custom_settings):
    """ sets up a SQLAlchemy engine and returns a connection to the database.
    The connection string used for testing can be specified via the
    ``KOTTI_TEST_DB_STRING`` environment variable.  The ``custom_settings``
    fixture is needed to allow users to import their models easily instead of
    having to override the ``connection``.
    """
    # the following setup is based on `kotti.resources.initialize_sql`,
    # except that it explicitly binds the session to a specific connection
    # enabling us to use savepoints independent from the orm, thus allowing
    # to `rollback` after using `transaction.commit`...
    from sqlalchemy import create_engine
    from kotti import DBSession
    from kotti import metadata
    from kotti.resources import _adjust_for_engine
    from kotti.testing import testing_db_url
    engine = create_engine(testing_db_url())
    _adjust_for_engine(engine)
    connection = engine.connect()
    DBSession.registry.clear()
    DBSession.configure(bind=connection)
    metadata.bind = engine
    return connection


@fixture(scope='session')
def content(connection, settings):
    """ sets up some default content using Kotti's testing populator.
    """
    from transaction import commit
    from kotti import metadata
    from kotti.resources import get_root
    metadata.drop_all(connection.engine)
    metadata.create_all(connection.engine)
    # to create the default content with the correct workflow state
    # the workflow must be initialized first;  please note that these
    # settings won't persist, though;  use the `workflow` fixture if needed
    from zope.configuration import xmlconfig
    import kotti
    xmlconfig.file('workflow.zcml', kotti, execute=True)
    for populate in settings['kotti.populators']:
        populate()

    # We set the path here since it's required for some integration
    # tests, and because the 'content' fixture does not depend on
    # 'event' and therefore the event handlers aren't fired for root
    # otherwise:
    get_root().path = u'/'
    commit()


@fixture
def db_session(config, content, connection, request):
    """ returns a db session object and sets up a db transaction
        savepoint, which will be rolled back after the test.
    """

    from transaction import abort
    trans = connection.begin()          # begin a non-orm transaction
    request.addfinalizer(trans.rollback)
    request.addfinalizer(abort)
    from kotti import DBSession
    return DBSession()


@fixture
def dummy_request(config, request, monkeypatch):
    """ returns a dummy request object after registering it as
        the currently active request.  This is needed when
        `pyramid.threadlocal.get_current_request` is used.
    """

    from kotti.testing import DummyRequest

    if 'user' in request.keywords:
        monkeypatch.setattr(
            DummyRequest,
            "authenticated_userid",
            request.keywords['user'].args[0])

    config.manager.get()['request'] = dummy_request = DummyRequest()

    return dummy_request


@fixture
def dummy_mailer(monkeypatch):
    from pyramid_mailer.mailer import DummyMailer

    mailer = DummyMailer()
    monkeypatch.setattr('kotti.message.get_mailer', lambda: mailer)
    return mailer


@fixture
def events(config, request):
    """ sets up Kotti's default event handlers.
    """
    from kotti.events import clear
    config.include('kotti.events')
    request.addfinalizer(clear)
    return config


@fixture
def setup_app(unresolved_settings, filedepot):
    from kotti import base_configure
    config = base_configure({}, **unresolved_settings)
    return config.make_wsgi_app()


@fixture
def app(workflow, db_session, dummy_mailer, events, setup_app):
    return setup_app


@fixture
def browser(db_session, request, setup_app):
    """ returns an instance of `zope.testbrowser`.  The `kotti.testing.user`
        pytest marker (or `pytest.mark.user`) can be used to pre-authenticate
        the browser with the given login name: `@user('admin')`.
    """
    from zope.testbrowser.wsgi import Browser
    from kotti.testing import BASE_URL
    host, port = BASE_URL.split(':')[-2:]
    browser = Browser('http://{}:{}/'.format(host[2:], int(port)),
                      wsgi_app=setup_app)
    if 'user' in request.keywords:
        # set auth cookie directly on the browser instance...
        from pyramid.security import remember
        from pyramid.testing import DummyRequest
        login = request.keywords['user'].args[0]
        environ = dict(HTTP_HOST=host[2:])
        for _, value in remember(DummyRequest(environ=environ), login):
            cookie, _ = value.split(';', 1)
            name, value = cookie.split('=')
            if name in browser.cookies:
                del browser.cookies[name]
            browser.cookies.create(name, value.strip('"'), path='/')
    return browser


@fixture
def root(db_session):
    """ returns Kotti's 'root' node.
    """
    from kotti.resources import get_root
    return get_root()


@fixture
def webtest(app, monkeypatch, request, filedepot, dummy_mailer):
    from webtest import TestApp
    if 'user' in request.keywords:
        login = request.keywords['user'].args[0]
        monkeypatch.setattr(
            "pyramid.authentication."
            "AuthTktAuthenticationPolicy.unauthenticated_userid",
            lambda self, req: login)
    return TestApp(app)


@fixture
def workflow(config):
    """ loads and activates Kotti's default workflow rules.
    """
    from zope.configuration import xmlconfig
    import kotti
    xmlconfig.file('workflow.zcml', kotti, execute=True)


class TestStorage:
    def __init__(self):
        self._storage = {}
        self._storage.setdefault(0)

    def get(self, id):
        info = self._storage[int(id)]

        from StringIO import StringIO

        f = MagicMock(wraps=StringIO(info['content']))
        f.seek(0)
        f.public_url = None
        f.filename = info['filename']
        f.content_type = info['content_type']
        f.content_length = len(info['content'])
        # needed to make JSON serializable, Mock objects are not
        f.last_modified = datetime(2012, 12, 30)

        return f

    def create(self, content, filename=None, content_type=None):
        _id = max(self._storage) + 1
        filename = filename or getattr(content, 'filename', None)
        content_type = content_type or getattr(content, 'type', None)

        if hasattr(content, 'file') and hasattr(content.file, 'read'):
            content = content.file.read()
        elif hasattr(content, 'read'):
            content = content.read()

        self._storage[_id] = {'content': content,
                              'filename': filename,
                              'content_type': content_type}
        return _id

    def delete(self, id):
        del self._storage[int(id)]


@fixture
def depot_tween(request, config, dummy_request):
    """ Sets up the Depot tween and patches Depot's ``set_middleware`` to
    suppress exceptions on subsequent calls """

    from depot.manager import DepotManager
    from kotti.filedepot import TweenFactory
    from kotti.filedepot import uploaded_file_response
    from kotti.filedepot import uploaded_file_url

    dummy_request.__class__.uploaded_file_response = uploaded_file_response
    dummy_request.__class__.uploaded_file_url = uploaded_file_url

    _set_middleware = DepotManager.set_middleware
    TweenFactory(None, config.registry)

    # noinspection PyDecorator
    @classmethod
    def set_middleware_patched(cls, mw):
        pass

    DepotManager.set_middleware = set_middleware_patched

    def restore():
        DepotManager.set_middleware = _set_middleware

    request.addfinalizer(restore)


@fixture
def mock_filedepot(request, depot_tween):
    """ Configures a mock depot store for :class:`depot.manager.DepotManager`

    This filedepot is not integrated with dbsession.
    Can be used in simple, standalone unit tests.
    """
    from depot.manager import DepotManager

    DepotManager._depots = {
        'mockdepot': MagicMock(wraps=TestStorage())
    }
    DepotManager._default_depot = 'mockdepot'

    def restore():
        DepotManager._clear()

    request.addfinalizer(restore)


@fixture
def filedepot(db_session, request, depot_tween):
    """ Configures a dbsession integrated mock depot store for
    :class:`depot.manager.DepotManager`
    """
    from depot.manager import DepotManager

    DepotManager._depots = {
        'filedepot': MagicMock(wraps=TestStorage())
    }
    DepotManager._default_depot = 'filedepot'

    def restore():
        db_session.rollback()
        DepotManager._clear()

    request.addfinalizer(restore)


@fixture
def no_filedepots(db_session, request, depot_tween):
    """ A filedepot fixture to empty and then restore DepotManager configuration
    """
    from depot.manager import DepotManager

    DepotManager._depots = {}
    DepotManager._default_depot = None

    def restore():
        db_session.rollback()
        DepotManager._clear()

    request.addfinalizer(restore)



# -*- coding: utf-8 -*-

unsanitized = u'''
<h1>Title</h1>
<div class="teaser umlaut">Descrüptiön</div>
<p>
  Paragraph with
  <a href="internal.html" target="_blank">internal</a> and
  <a href="http://external.com/" target="_blank">external</a> links.
</p>
<marquee>Fancy marquee!</marquee>
<b SIZE=17 style="color: red">IMPORTANT!</b>
<script>
  alert('XSS!')
</script>
<p>Unclosed paragraph
'''


def _verify_no_html(sanitized):
    assert u'<' not in sanitized
    assert u'external links' in sanitized


def test_no_html():

    from kotti.sanitizers import no_html

    _verify_no_html(no_html(unsanitized))


def _verify_minimal_html(sanitized):

    from bleach_whitelist import all_tags
    from bleach_whitelist import markdown_tags
    from bleach_whitelist import print_tags

    for tag in set(all_tags) - set(markdown_tags) - set(print_tags):
        assert u'<{0}'.format(tag) not in sanitized

    assert u'style=""' in sanitized
    assert u'<a href="http://external.com/">' in sanitized
    assert u'size' not in sanitized.lower()


def test_minmal_html():

    from kotti.sanitizers import minimal_html

    _verify_minimal_html(minimal_html(unsanitized))


def _verify_xss_protection(sanitized):

    assert u'<script>' not in sanitized
    assert u'<h1>Title</h1>' in sanitized
    assert u'<a href="internal.html" target="_blank">internal</a>' in sanitized
    assert u'b size="17"' in sanitized
    assert u'<p>Unclosed paragraph\n</p>' in sanitized


def test_xss_protection():

    from kotti.sanitizers import xss_protection

    _verify_xss_protection(xss_protection(unsanitized))


def test_default_config(unresolved_settings):

    assert u'kotti.sanitizers' in unresolved_settings
    assert u'kotti.sanitize_on_write' in unresolved_settings

    assert unresolved_settings[u'kotti.sanitizers'] == u'xss_protection:kotti.sanitizers.xss_protection minimal_html:kotti.sanitizers.minimal_html no_html:kotti.sanitizers.no_html'  # noqa
    assert unresolved_settings[u'kotti.sanitize_on_write'] == u'kotti.resources.Document.body:xss_protection kotti.resources.Content.title:no_html kotti.resources.Content.description:no_html'  # noqa


def test_setup_sanitizers(unresolved_settings):
    from kotti.sanitizers import _setup_sanitizers
    from kotti.sanitizers import minimal_html
    from kotti.sanitizers import no_html
    from kotti.sanitizers import xss_protection

    _setup_sanitizers(unresolved_settings)

    settings = unresolved_settings[u'kotti.sanitizers']

    assert u'minimal_html' in settings
    assert u'no_html' in settings
    assert u'xss_protection' in settings

    assert settings[u'minimal_html'] == minimal_html
    assert settings[u'no_html'] == no_html
    assert settings[u'xss_protection'] == xss_protection


def test_listeners(app, root, db_session):

    from kotti.resources import Document

    root[u'd'] = doc = Document(
        name=u'test',
        title=u'<h1>Title</h1>',
        description=unsanitized,
        body=unsanitized)

    db_session.flush()

    assert doc.title == u'Title'
    _verify_no_html(doc.description)
    _verify_xss_protection(doc.body)


def test_sanitize(app, dummy_request):

    from kotti.sanitizers import sanitize
    from kotti.resources import Document
    from kotti.views.util import TemplateAPI

    _verify_no_html(sanitize(unsanitized, 'no_html'))
    _verify_minimal_html(sanitize(unsanitized, 'minimal_html'))
    _verify_xss_protection(sanitize(unsanitized, 'xss_protection'))

    api = TemplateAPI(Document(), dummy_request)
    _verify_no_html(api.sanitize(unsanitized, 'no_html'))
    _verify_minimal_html(api.sanitize(unsanitized, 'minimal_html'))
    _verify_xss_protection(api.sanitize(unsanitized, 'xss_protection'))



# coding:utf8

import pytest
from kotti.testing import BASE_URL
from kotti.testing import user
from mock import patch
from webtest.forms import Upload


class TestLogin:
    def test_it(self, app):
        res = app.post(
            '/@@login', dict(login='admin', password='secret', submit='submit'))
        assert res.status == '302 Found'
        res = res.maybe_follow()
        assert res.status == '200 OK'


class TestForbidden:
    def test_forbidden(self, app):
        app.get('/@@edit', headers={'Accept': '*/json'}, status=403)

    def test_forbidden_redirect(self, app):
        res = app.get('/@@edit', headers={'Accept': 'text/html'}, status=302)
        assert res.location.startswith('http://localhost/@@login?came_from=')

    def test_forbidden_redirect_when_authenticated(self, app):
        with patch('pyramid.request.Request.authenticated_userid', 'foo'):
            res = app.get('/@@edit', status=302)
        assert res.location == 'http://localhost/@@forbidden'


class TestUploadFile:

    def add_file(self, browser, contents='ABC'):
        file_ctrl = browser.getControl("File")
        file_ctrl.add_file(contents, 'image/gif', 'my_image.gif')
        browser.getControl('save').click()

    @user('admin')
    def test_it(self, browser, filedepot):
        browser.open(BASE_URL + '/@@add_file')
        self.add_file(browser)
        assert "Item was added" in browser.contents

    @user('admin')
    def test_view_uploaded_file(self, browser, filedepot):
        browser.open(BASE_URL + '/@@add_file')
        self.add_file(browser)
        browser.getLink("View").click()
        browser.getLink("Download file").click()
        assert browser.contents == 'ABC'

    @user('admin')
    def test_tempstorage(self, browser, filedepot):
        browser.open(BASE_URL + '/@@add_file')
        self.add_file(browser, contents='DEF')
        browser.getLink("Edit").click()
        browser.getControl("Title").value = ''  # the error
        assert "Your changes have been saved" not in browser.contents
        browser.getControl("Title").value = 'A title'
        browser.getControl('save').click()
        assert "Your changes have been saved" in browser.contents
        browser.getLink("View").click()
        browser.getLink("Download file").click()
        assert browser.contents == 'DEF'

    @user('admin')
    def test_edit_uploaded_file(self, browser, filedepot):
        browser.open(BASE_URL + '/@@add_file')
        self.add_file(browser, contents='GHI')
        browser.getLink("Edit").click()
        browser.getControl('save').click()
        browser.getLink("Download file").click()
        assert browser.contents == 'GHI'


class TestValidatorMaxLength:

    @user('admin')
    def test_title_max_length_document_ko(self, browser, filedepot):
        browser.open(BASE_URL + '/@@add_document')
        from kotti.resources import Node
        max_length = Node.name.property.columns[0].type.length
        browser.getControl("Title").value = '1' * (max_length + 1)  # the error
        browser.getControl('save').click()
        assert "Item was added" not in browser.contents

    @user('admin')
    def test_title_max_length_document_ok(self, browser):
        browser.open(BASE_URL + '/@@add_document')
        from kotti.resources import Node
        max_length = Node.name.property.columns[0].type.length
        browser.getControl("Title").value = '1' * max_length
        browser.getControl('save').click()
        assert "Item was added" in browser.contents

    @user('admin')
    def test_title_max_length_file_ok(self, browser):
        browser.open(BASE_URL + '/@@add_file')
        from kotti.resources import Node
        max_length = Node.name.property.columns[0].type.length
        browser.getControl("Title").value = '1' * max_length
        file_ctrl = browser.getControl("File")
        file_ctrl.add_file("abc", 'image/gif', 'my_image.gif')
        browser.getControl('save').click()
        assert "Item was added" in browser.contents

    @user('admin')
    def test_title_max_length_file_ko(self, browser):
        browser.open(BASE_URL + '/@@add_file')
        from kotti.resources import Node
        max_length = Node.name.property.columns[0].type.length
        browser.getControl("Title").value = '1' * (max_length + 1)  # the error
        file_ctrl = browser.getControl("File")
        file_ctrl.add_file("abc", 'image/gif', 'my_image.gif')
        browser.getControl('save').click()
        assert "Item was added" not in browser.contents


class TestBrowser:
    """ This is a one to one conversion of the former browsert.txt.
    These tests should definitively be rewritten and splitted into multiple
    methods to ease readability.
    """

    @staticmethod
    def _login(app, login, password):
        resp = app.get('/@@login')
        form = resp.forms['login-form']
        form['login'] = login
        form['password'] = password
        resp = form.submit('submit')
        return resp

    @staticmethod
    def _select_children(resp, *child_idx):
        """ Mark the checkbox(es) of the rows in the ``contents`` view

        :param resp: response which's body contains the contents table
        :type resp: :class:`webtest.response.TestResponse`

        :param *child_idx: index of the child in the folder listing
                           (starting with 0)
        :type *child_idx: int

        :return: contents form with children selected (ready for submission)
        :rtype: :class:`webtest.forms.Form`
        """

        form = resp.forms['form-contents']
        form['children'] = [
            cb._value for cb in [
                form.fields['children'][idx] for idx in child_idx]]

        return form

    def test_login(self, webtest):

        # get the frontpage
        resp = webtest.app.get('/')
        assert "You have successfully installed Kotti." in resp.body

        # access a protected URL
        resp = webtest.app.get('/edit').maybe_follow()
        assert "Log in" in resp.body

        # submit the login form
        form = resp.forms['login-form']
        assert form['came_from'].value == 'http://localhost/edit'
        form['login'] = 'admin'
        form['password'] = 'secret'

        resp = form.submit('submit').maybe_follow()
        assert "Welcome, Administrator" in resp.body

        # log out
        resp = resp.click('Logout').maybe_follow()
        assert "You have been logged out" in resp.body

        # attempt to login with wrong credentials
        form = resp.forms['login-form']
        assert form['came_from'].value == 'http://localhost/edit'
        form['login'] = 'admin'
        form['password'] = 'funny'

        resp = form.submit('submit')
        assert "Welcome, Adminstrator" not in resp.body
        assert "Login failed" in resp.body

        # and again, this time with correct credentials
        form = resp.forms['login-form']
        assert form['came_from'].value == 'http://localhost/edit'
        form['login'] = 'admin'
        form['password'] = 'secret'
        resp = form.submit('submit').maybe_follow()
        assert "Welcome, Administrator" in resp.body

    @pytest.mark.user('admin')
    def test_content_management(self, webtest):

        from kotti.resources import Document
        from kotti.resources import File
        from kotti_image.resources import Image

        save_addable_document = Document.type_info.addable_to
        save_addable_file = File.type_info.addable_to
        save_addable_image = Image.type_info.addable_to

        app = webtest.app
        resp = app.get('/')

        # Add a document
        resp = resp.click('Document', index=0)
        assert "Add Document to Welcome to Kotti" in resp.body
        form = resp.forms['deform']
        form['title'] = "Child One"
        resp = form.submit('save').maybe_follow()
        assert "Item was added" in resp.body
        assert resp.request.path == '/child-one/'

        # Edit the document
        resp = resp.click('Edit')
        assert "Edit Child One" in resp.body
        form = resp.forms['deform']
        form['title'] = "First Child"
        resp = form.submit('save').maybe_follow()
        assert "Your changes have been saved" in resp.body
        resp = resp.click('Edit')
        form = resp.forms['deform']
        assert form['title'].value == "First Child"
        resp = resp.click('Edit')
        assert "First Child" in resp.body

        # And now force a validation error:
        resp = resp.click('Edit')
        form = resp.forms['deform']
        form['title'] = ""
        resp = form.submit('save')
        assert "There was a problem" in resp.body
        assert form['title'].value == ''

        # And now click the 'Cancel' button:
        resp = resp.click('Edit')
        form = resp.forms['deform']
        form['title'] = "Firstborn"
        resp = form.submit('cancel').maybe_follow()
        assert 'deform' not in resp.forms
        assert "Firstborn" not in resp.body

        # Now click the 'Cancel' button for an invalid form entry:
        resp = resp.click('Edit')
        form = resp.forms['deform']
        form['title'] = ""
        resp = form.submit('cancel').maybe_follow()
        assert 'deform' not in resp.forms
        assert "Firstborn" not in resp.body

        # Add two more documents, at different levels
        resp = app.get('/')
        resp = resp.click('Document', index=0)
        form = resp.forms['deform']
        form['title'] = "Second Child"
        resp = form.submit('save').maybe_follow()
        assert "Item was added" in resp.body
        assert resp.request.path == '/second-child/'

        resp = resp.click('Document', index=0)
        form = resp.forms['deform']
        form['title'] = "Grandchild"
        resp = form.submit('save').maybe_follow()
        assert "Item was added" in resp.body
        assert resp.request.path == '/second-child/grandchild/'

        # Add another grandchild with the same name:
        resp = app.get('/second-child/')
        resp = resp.click('Document', index=0)
        form = resp.forms['deform']
        form['title'] = "Grandchild"
        resp = form.submit('save').maybe_follow()
        assert "Item was added" in resp.body
        assert resp.request.path == '/second-child/grandchild-1/'

        # There's no Add link if nothing can be added:
        resp = app.get('/second-child/grandchild-1/')
        try:
            Document.type_info.addable_to = ()
            File.type_info.addable_to = ()
            Image.type_info.addable_to = ()
            with pytest.raises(IndexError):
                resp.click(href='add_')
        finally:
            Document.type_info.addable_to = save_addable_document
            File.type_info.addable_to = save_addable_file
            Image.type_info.addable_to = save_addable_image

        # Add a file
        resp = app.get('/')
        resp = resp.click('File', index=0)
        form = resp.forms['deform']
        form['description'] = "A tiny file"
        form['upload'] = Upload('tiny.txt', 'tiny', 'text/plain')
        resp = form.submit('save').maybe_follow()
        assert "Item was added" in resp.body
        assert resp.request.path == '/tiny.txt/'

        # Add a file larger than maximum file size
        resp = app.get('/')
        resp = resp.click('File', index=0)
        form = resp.forms['deform']
        form['title'] = "Huge file"
        form['description'] = "An uge file"
        form['upload'] = Upload('huge.txt',
                                '*' * (10 * 1024 * 1024 + 1), 'text/plain')
        resp = form.submit('save')
        assert "There was a problem" in resp.body
        assert "Maximum file size" in resp.body

        # Add tags to a document:
        resp = app.get('/second-child/')
        resp = resp.click('Document', index=0)
        form = resp.forms['deform']
        form['title'] = "Grandchild"
        form['tags'] = ''
        form.submit('save').maybe_follow()
        resp = app.get('/second-child/')
        resp = resp.click('Document', index=0)
        form = resp.forms['deform']
        form['title'] = "Grandchild"
        form['tags'] = 'tag 1, tag 2,tag 3'
        form.submit('save').maybe_follow()
        resp = app.get('/second-child/grandchild-2/@@edit')
        form = resp.forms['deform']
        assert 'tag 1' in resp.body
        assert 'tag 2' in resp.body
        assert 'tag 3' in resp.body
        form['tags'] = 'tag 1, tag 4, tag 5,tag 6, tag 7, übertag'
        form.submit('save').maybe_follow()
        resp = app.get('/second-child/grandchild-2/@@edit')
        assert 'value="tag 1,tag 4,tag 5,tag 6,tag 7,übertag"' in resp.body

        # Delete a document
        resp = app.get('/second-child/grandchild/')
        resp = resp.click('Delete', index=0)
        form = resp.forms['form-delete']
        resp = form.submit('cancel')
        assert "Grandchild was deleted" not in resp.body
        resp = app.get('/second-child/grandchild/')
        resp = resp.click('Delete', index=0)
        form = resp.forms['form-delete']
        resp = form.submit('delete', value='delete').maybe_follow()
        assert "Grandchild was deleted" in resp.body
        assert resp.request.path == '/second-child/'

        # Copy and paste
        resp = app.get('/second-child/')
        resp = resp.click('Cut', index=0).maybe_follow()
        assert "Second Child was cut" in resp.body
        resp = app.get('/child-one/')
        resp = resp.click('Paste', index=0).maybe_follow()
        assert "Second Child was pasted" in resp.body
        app.get('/second-child/', status=404)
        resp = app.get('/child-one/second-child/')
        resp = resp.click('Copy', index=0).maybe_follow()
        assert "Second Child was copied" in resp.body
        resp = app.get('/')
        resp = resp.click('Paste', index=0).maybe_follow()
        assert "Second Child was pasted" in resp.body

        # We can paste twice since we copied:
        resp = app.get('/')
        resp = resp.click('Paste', index=0).maybe_follow()
        assert "Second Child was pasted" in resp.body
        resp = app.get('/second-child/')
        assert "Second Child" in resp.body
        resp = app.get('/second-child-1/')
        assert "Second Child" in resp.body

        # We can also copy and paste items that contain children,
        # like the whole site:
        resp = app.get('/')
        resp = resp.click('Copy', index=0).maybe_follow()
        assert "Welcome to Kotti was copied" in resp.body
        resp = app.get('/second-child/')
        resp = resp.click('Paste', index=0).maybe_follow()
        assert "Welcome to Kotti was pasted" in resp.body
        resp = app.get('/second-child/welcome-to-kotti/')
        assert resp.status_code == 200
        resp = app.get('/second-child/welcome-to-kotti/second-child/')
        assert resp.status_code == 200

        # And finally cut and paste a tree:
        resp = app.get('/second-child/')
        resp.click('Cut', index=0).maybe_follow()
        resp = app.get('/child-one/second-child/')
        resp = resp.click('Paste', index=0).maybe_follow()
        assert "Second Child was pasted" in resp.body
        app.get('/second-child/', status=404)

        # Note how we can't cut and paste an item into itself:
        resp = app.get('/child-one/')
        resp.click('Cut', index=0).maybe_follow()
        with pytest.raises(IndexError):
            resp.click('Paste', index=0).maybe_follow()
        resp = app.get('/child-one/second-child/')
        with pytest.raises(IndexError):
            resp.click('Paste', index=0).maybe_follow()

        # Whether we can paste or not also depends on the
        # ``type_info.addable`` property:
        resp = app.get('/child-one/')
        resp.click('Copy', index=0).maybe_follow()
        resp = app.get('/child-one/second-child/')
        resp.click('Paste', index=0).maybe_follow()
        try:
            Document.type_info.addable_to = ()
            resp = app.get('/child-one/second-child/')
            with pytest.raises(IndexError):
                resp.click('Paste', index=0).maybe_follow()
        finally:
            Document.type_info.addable_to = save_addable_document

        # You can't cut the root of a site:
        resp = app.get('/child-one/')
        resp.click('Cut', index=0)
        resp = app.get('/')
        with pytest.raises(IndexError):
            resp.click('Cut', index=0)

        # We can rename an item. Slashes will be stripped out.:
        resp = app.get('/child-one/second-child/')
        resp = resp.click('Rename', index=0)
        form = resp.forms['form-rename']
        assert form['name'].value == 'second-child'
        assert form['title'].value == 'Second Child'
        form['name'] = 'thi/rd-ch/ild'
        form['title'] = 'My Third Child'
        resp = form.submit('rename').maybe_follow()
        assert "Item was renamed" in resp.body
        assert resp.request.path == '/child-one/third-child/'

        # We cannot rename the root:
        resp = app.get('/')
        with pytest.raises(IndexError):
            resp.click('Rename', index=0)

        # On setup pages we can't use the actions:
        resp = app.get('/')
        resp = resp.click('User Management', index=0)
        with pytest.raises(IndexError):
            resp.click('Copy', index=0)
        resp = resp.click('Preferences', index=0)
        with pytest.raises(IndexError):
            resp.click('Copy', index=0)

        # Contents view actions
        resp = app.get('/child-one')
        resp = resp.click('Contents', index=0)
        assert resp.request.path == '/child-one/@@contents'
        resp = resp.forms['form-contents'].submit('copy').maybe_follow()
        assert 'You have to select items' in resp.body
        form = self._select_children(resp, 0)
        resp = form.submit('copy').maybe_follow()
        assert 'My Third Child was copied.' in resp.body
        resp = app.get('/second-child-1/@@contents')
        assert 'My Third Child' not in resp.body
        resp = resp.forms['form-contents'].submit('paste').maybe_follow()
        assert 'My Third Child' in resp.body
        resp = app.get('/second-child-1/@@contents')
        form = self._select_children(resp, 0)
        resp = form.submit('cut').maybe_follow()
        assert 'cut.' in resp.body
        resp = app.get('/child-one/@@contents')
        assert "Grandchild" not in resp.body
        resp = resp.forms['form-contents'].submit('paste').maybe_follow()
        assert "Grandchild" in resp.body
        with pytest.raises(IndexError):
            resp.click('Paste')
        resp = app.get('/child-one/@@contents')
        form = self._select_children(resp, 0, 1)
        form.submit('cut').maybe_follow()
        resp = app.get('/')
        resp = resp.click('Document', index=0)
        form = resp.forms['deform']
        form['title'] = 'Forth child'
        form.submit('save').maybe_follow()
        resp = app.get('/forth-child/@@contents')
        assert "Grandchild" not in resp.body
        assert "My Third Child" not in resp.body
        resp.forms['form-contents'].submit('paste').maybe_follow()
        resp = app.get('/forth-child/@@contents')
        assert "Grandchild" in resp.body
        assert "My Third Child" in resp.body
        resp = app.get('/child-one/@@contents')
        assert "Grandchild" not in resp.body
        resp = app.get('/forth-child/@@contents')
        assert 'third-child' in resp.body
        assert 'Grandchild' in resp.body
        assert 'child-the-third' not in resp.body
        assert 'Hello Bob' not in resp.body
        form = self._select_children(resp, 0, 1)
        resp = form.submit('rename_nodes').maybe_follow()
        resp = resp.forms['form-rename-nodes'].submit('cancel').maybe_follow()
        assert 'No changes were made.' in resp.body
        assert resp.request.path == '/forth-child/@@contents'
        form = self._select_children(resp, 0, 1)
        resp = form.submit('rename_nodes').maybe_follow()
        form = resp.forms['form-rename-nodes']
        form[form.submit_fields()[1][0]] = 'child-the-third'
        form[form.submit_fields()[2][0]] = 'child, the third'
        form[form.submit_fields()[4][0]] = 'hello-bob'
        form[form.submit_fields()[5][0]] = 'Hello Bob'
        resp = form.submit('rename_nodes').maybe_follow()
        assert resp.request.path == '/forth-child/@@contents'
        assert 'third-child' not in resp.body
        assert 'Grandchild' not in resp.body
        assert 'child-the-third' in resp.body
        assert 'Hello Bob' in resp.body

        resp = resp.click('File', index=0)
        form = resp.forms['deform']
        form['description'] = 'A file'
        form['upload'] = Upload('some.txt', 'something', 'text/plain')
        form.submit('save').maybe_follow()
        resp = app.get('/forth-child/@@contents')
        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit('delete_nodes').maybe_follow()
        resp = resp.forms['form-delete-nodes'].submit('cancel').maybe_follow()
        assert 'No changes were made.' in resp.body
        assert resp.request.path == '/forth-child/@@contents'
        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit('delete_nodes').maybe_follow()
        assert "Are you sure" in resp.body
        resp = resp.forms['form-delete-nodes'].submit('delete_nodes',
                                                      status=302).maybe_follow()
        assert "child, the third was deleted." in resp.body
        assert "Hello Bob was deleted." in resp.body
        assert "some.txt was deleted." in resp.body
        resp = app.get('/forth-child/@@contents')
        assert "Welcome to Kotti" in resp.body
        assert '<i class="glyphicon glyphicon-home"></i>' in resp.body
        assert '<i class="glyphicon glyphicon-folder-open"></i>' in resp.body
        assert '<i class="glyphicon glyphicon-folder-close"></i>' not in resp.body  # noqa

        # Contents view change state actions
        resp = resp.click("Second Child")
        resp = resp.click("Contents")
        assert '/second-child-1/third-child/@@workflow-change?new_state=public' in resp.body  # noqa
        resp = app.get('/second-child-1/third-child/@@workflow-change?new_state=public').maybe_follow()  # noqa
        assert '/second-child-1/third-child/@@workflow-change?new_state=private' in resp.body  # noqa
        resp = app.get('/second-child-1/third-child/@@contents')
        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit('change_state').maybe_follow()
        assert 'Change workflow state' in resp.body
        resp = resp.forms['form-change-state'].submit('cancel').maybe_follow()
        assert 'No changes were made.' in resp.body
        assert resp.request.path == '/second-child-1/third-child/@@contents'
        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit('change_state').maybe_follow()
        form = resp.forms['form-change-state']
        form['children-to-change-state'] = []
        resp = form.submit('change_state').maybe_follow()
        assert 'No changes were made.' in resp.body
        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit('change_state').maybe_follow()
        form = resp.forms['form-change-state']
        form['to-state'] = 'public'
        resp = form.submit('change_state').maybe_follow()
        assert 'Your changes have been saved.' in resp.body
        assert '/second-child-1/third-child/grandchild-1/@@workflow-change?new_state=private' in resp.body  # noqa
        assert '/second-child-1/third-child/grandchild-2/@@workflow-change?new_state=private' in resp.body  # noqa
        assert '/second-child-1/third-child/grandchild-3/@@workflow-change?new_state=private' in resp.body  # noqa

        resp = resp.click('My Third Child')
        assert '/second-child-1/third-child/child-one/@@workflow-change?new_state=public' in resp.body  # noqa
        app.get('/second-child-1/third-child/child-one/@@workflow-change?new_state=public')  # noqa
        resp = app.get('/second-child-1/third-child/@@contents')
        assert '/second-child-1/third-child/child-one/@@workflow-change?new_state=private' in resp.body  # noqa
        resp = resp.click('First Child', index=1)
        resp = resp.click('Document', index=0)
        form = resp.forms['deform']
        form['title'] = 'Sub child'
        resp = form.submit('save').maybe_follow()
        assert '/second-child-1/third-child/child-one/sub-child/@@workflow-change?new_state=public' in resp.body  # noqa
        resp = resp.click("Second Child", index=0)
        resp = resp.click("Contents")
        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit('change_state').maybe_follow()
        form = resp.forms['form-change-state']
        form['include-children'] = 'include-children'
        form['to-state'] = 'public'
        resp = form.submit('change_state').maybe_follow()
        assert 'Your changes have been saved.' in resp.body
        assert '/second-child-1/third-child/@@workflow-change?new_state=private' in resp.body
        resp = app.get('/second-child-1/third-child/child-one/sub-child/')
        assert '/second-child-1/third-child/child-one/sub-child/@@workflow-change?new_state=private' in resp.body

        # Navigation
        resp = resp.click("Navigate")
        resp = resp.click("Second Child", index=0)
        assert resp.request.path == '/second-child-1/'

    def test_user_management(self, webtest, settings, dummy_mailer):
        from kotti import get_settings
        get_settings()['kotti.site_title'] = u'Website des Kottbusser Tors'

        app = webtest.app
        resp = self._login(app, 'admin', 'secret').maybe_follow()

        # The user management screen is available through
        # the "Site Setup" submenu
        resp = resp.click('User Management')

        # Add Bob's Group and assign the ``viewer`` role
        form = resp.forms['deform_group_add']
        form['name'] = 'bobsgroup'
        form['title'] = "Bob's Group"
        form.get('checkbox', index=0).checked = True
        resp = form.submit('add_group').maybe_follow()
        assert "Bob's Group was added" in resp.body
        assert resp.forms['form-global-roles'][
            'role::group:bobsgroup::role:viewer'].value == 'on'

        # And a Bob.
        # Only alphanumeric characters are allowed for the name
        form = resp.forms['deform_user_add']
        form["name"] = "bob:"
        form["title"] = "Bob Dabolina"
        form["password"] = "secret"
        form['password-confirm'] = "secret"
        form["email"] = "bob@DABOLINA.com"
        resp = form.submit("add_user")
        assert "There was a problem" in resp.body

        # Use a valid username now.  Note how the checkbox to send a password
        # registration link is ticked:
        form = resp.forms['deform_user_add']
        form["name"] = "Bob"
        form["email"] = "bob@DABOLINA.com"
        form["send_email"].checked = True
        resp = form.submit("add_user").maybe_follow()
        assert "Bob Dabolina was added" in resp.body

        # We cannot add Bob twice:
        form = resp.forms['deform_user_add']
        form["name"] = "bob"
        form["title"] = "Bobby Brown"
        form["password"] = "secret"
        form["password-confirm"] = "secret"
        resp = form.submit("add_user")
        assert "A user with that name already exists" in resp.body

        # We cannot add Bob Dabolina's email twice:
        form = resp.forms['deform_user_add']
        form["name"] = "bob2"
        form["title"] = "Bobby Brown"
        form["email"] = "bob@dabolina.com"
        form["password"] = "secret"
        form['password-confirm'] = "secret"
        resp = form.submit("add_user")
        assert "A user with that email already exists" in resp.body
        form = resp.forms['deform_user_add']
        form["email"] = "bob@gmail.com"
        resp = form.submit("add_user").maybe_follow()
        assert "Bobby Brown was added" in resp.body

        # Searching for Bob will return both Bob and Bob's Group:
        form = resp.forms['form-principal-search']
        form["query"] = "Bob"
        resp = form.submit("search")
        assert "Bob Dabolina" in resp.body
        assert "Bob's Group" in resp.body
        form = resp.forms['form-global-roles']
        assert form["role::group:bobsgroup::role:viewer"].checked
        assert not form["role::group:bobsgroup::role:editor"].checked
        assert not form["role::bob::role:editor"].checked

        # We can click on the Bob's entry to edit the list of groups
        # he belongs to:
        resp = resp.click("Bob Dabolina")
        form = resp.forms['deform']
        form["group"] = "bobsgroup"
        resp = form.submit("save").maybe_follow()
        assert "Your changes have been saved" in resp.body

        # We cannot update on the Bobby Brown's entry with duplicated email:
        resp = resp.click("Back to User Management")
        form = resp.forms['form-principal-search']
        form["query"] = "Bobby"
        resp = form.submit("search")
        resp = resp.click("Bobby Brown")
        form = resp.forms['deform']
        form["email"] = "bob@dabolina.com"
        resp = form.submit("save")
        assert "A user with that email already exists" in resp.body

        # If we cancel the edit of the user, we are redirected to the
        # user management screen:
        resp = resp.forms['deform'].submit("cancel").maybe_follow()
        assert "No changes were made." in resp.body
        assert resp.request.path == '/@@setup-users'

        # Back in user management we can see how Bob's inherited the
        # viewer role from Bob's Group:
        form = resp.forms['form-principal-search']
        form["query"] = "Bob"
        resp = form.submit("search")
        form = resp.forms['form-global-roles']
        assert form["role::group:bobsgroup::role:viewer"].checked
        assert 'disabled' not in form[
            "role::group:bobsgroup::role:viewer"].attrs
        assert form["role::bob::role:viewer"].checked
        assert 'disabled' in form["role::bob::role:viewer"].attrs

        # We can click on the Bob's Group entry to edit an email address:
        resp = resp.click("Bob's Group")
        form = resp.forms['deform']
        form["email"] = 'bogsgroup@gmail.com'
        resp = form.submit("save").maybe_follow()
        assert "Your changes have been saved" in resp.body

        # Set password
        # ------------

        # Remember that we sent Bob an email for registration.
        # He can use it to set his own password:
        [email, email2] = dummy_mailer.outbox
        assert email.recipients == [u'"Bob Dabolina" <bob@dabolina.com>']
        assert email.subject == "Your registration for Website des " \
                                "Kottbusser Tors"
        assert "Hello, Bob Dabolina!" in email.body
        assert "You are joining Website des Kottbusser Tors." in email.body
        assert "Click here to set your password and log in:" in email.body
        assert "@@set-password?token=" in email.body

        # We'll use that link to set our password:
        resp.click("Logout").maybe_follow()
        path = email.body[email.body.index('http://localhost'):].split()[0][16:]
        resp = app.get(path)
        form = resp.forms['deform']
        form['password'] = "newpassword"
        form['password-confirm'] = "newpasswoops"  # a silly error
        resp = form.submit("submit")
        assert "There was a problem with your submission" in resp.body
        form = resp.forms['deform']
        form['password'] = "newpassword"
        form['password-confirm'] = "newpassword"
        resp = form.submit("submit").maybe_follow()
        assert "You have reset your password" in resp.body

        # We cannot use that link again:
        resp = app.get(path)
        form = resp.forms['deform']
        form['password'] = "won't work"
        form['password-confirm'] = "won't work"
        resp = form.submit("submit")
        assert "Your password reset token may have expired" in resp.body

        # Log in as Bob with the new password:
        resp = self._login(app, 'bOB', 'newpassword').maybe_follow()
        assert "Welcome, Bob Dabolina" in resp.body
        app.get('/@@edit')
        resp = resp.click("Logout").maybe_follow()

        # The login form has a "Reset password" button.  Let's try it:
        resp = self._login(app, "bobby", "")
        assert "Login failed." in resp.body
        form = resp.forms['forgot-password-form']
        form['login'] = "bob"
        resp = form.submit('reset-password').maybe_follow()
        assert "You should be receiving an email" in resp.body
        [email1, email2, email3] = dummy_mailer.outbox
        assert 'Hello, Bob Dabolina!' in email3.body
        assert 'Click this link to reset your password at Website des ' \
               'Kottbusser Tors:' in email3.body

        # User preferences
        # ----------------

        # The "Preferences" link leads us to a form where the user can change
        # their preferences so the user need to be authenticated:
        resp = self._login(app, 'admin', 'secret').maybe_follow()
        assert "Welcome, Administrator" in resp.body
        resp = app.get('/@@prefs')
        form = resp.forms['deform']
        form['title'] = "Mr. Administrator"
        form['email'] = 'admin@minad.com'
        resp = form.submit("save").maybe_follow()
        assert "Your changes have been saved" in resp.body

        # The email could not be used if already a users with that email exists:
        form = resp.forms['deform']
        form['email'] = 'bob@dabolina.com'
        resp = form.submit("save")

        # If the user cancel the process he will be redirected to the site root:
        form = resp.forms['deform']
        resp = form.submit("cancel").maybe_follow()
        assert 'Welcome to Kotti' in resp.body

        # Share
        # -----

        # The Share tab allows us to assign users and groups to roles:
        resp = app.get('/')
        resp = resp.click("Edit")
        resp = resp.click("Share")

        # We can search for users:
        form = resp.forms['form-share-search']
        form['query'] = "Bob"
        resp = form.submit("search")

        # Bob and Bob's Group are listed now:
        assert "Bob Dabolina" in resp.body
        assert "Bob's Group" in resp.body

        # We add Bob's Group to Owners and Editors before taking away Owners
        # again:
        form = resp.forms['form-share-assign']
        form["role::group:bobsgroup::role:owner"].checked = True
        form["role::group:bobsgroup::role:editor"].checked = True
        resp = form.submit("apply").maybe_follow()
        assert "Your changes have been saved" in resp.body
        resp = app.get(resp.request.path)  # >>> browser.reload()
        form = resp.forms['form-share-assign']
        assert form["role::group:bobsgroup::role:owner"].checked
        assert form["role::group:bobsgroup::role:editor"].checked
        form["role::group:bobsgroup::role:owner"].checked = False
        resp = form.submit("apply").maybe_follow()
        assert "Your changes have been saved" in resp.body
        form = resp.forms['form-share-assign']
        assert not form["role::group:bobsgroup::role:owner"].checked

        # Not making any changes will give us a different feedback message:
        resp = form.submit("apply")
        assert "Your changes have been saved" not in resp.body
        form = resp.forms['form-share-assign']
        assert not form["role::group:bobsgroup::role:owner"].checked
        assert form["role::group:bobsgroup::role:editor"].checked

        # Bob should now have an inherited Editor role, because he's part of
        # Bob's Group:
        form = resp.forms['form-share-search']
        form["query"] = "Bob Dabolina"
        resp = form.submit("search")
        form = resp.forms['form-share-assign']
        assert form["role::bob::role:editor"].checked
        # TODO assert 'disabled' in form["role::bob::role:editor"].attrs
        # TODO assert not form["role::bob::role:owner"].checked

        # Lastly, let's take away the remaining Editor role from Bob's Group
        # again:
        assert "Bob's Group" in resp.body
        form["role::group:bobsgroup::role:editor"].checked = None
        resp = form.submit("apply").maybe_follow()
        assert "Your changes have been saved" in resp.body
        # TODO assert "Bob's Group" not in resp.body
        form = resp.forms['form-share-assign']
        assert not form["role::group:bobsgroup::role:editor"].checked

        # Delete users
        # ------------

        resp = app.get('/')
        resp = resp.click("User Management")

        # We add a group and assign the ``Manager`` role to one:
        form = resp.forms['deform_group_add']
        form['name'] = "pelayogroup"
        form['title'] = "Pelayo Group"
        form.get('checkbox', index=3).checked = True
        resp = form.submit("add_group").maybe_follow()
        assert "Pelayo Group was added" in resp.body
        form = resp.forms['form-global-roles']
        assert form["role::group:pelayogroup::role:admin"].checked

        # And add some users to this group.
        form = resp.forms['deform_user_add']
        form['name'] = "Bruce"
        form['title'] = "Bruce Pelayo"
        form['password'] = "secret"
        form['password-confirm'] = "secret"
        form['email'] = "bruce@pelayno.com"
        form["group"] = "pelayogroup"
        resp = form.submit("add_user").maybe_follow()
        assert "Bruce Pelayo was added" in resp.body

        form = resp.forms['deform_user_add']
        form['name'] = "Brian"
        form['title'] = "Brian Pelayo"
        form['password'] = "secret"
        form['password-confirm'] = "secret"
        form['email'] = "brian@pelayno.com"
        form["group"] = "pelayogroup"
        resp = form.submit("add_user").maybe_follow()
        assert "Brian Pelayo was added" in resp.body

        # Lets login as Bruce and add some content:
        resp.click("Logout").maybe_follow()
        resp = self._login(app, 'bruce', 'secret').maybe_follow()
        assert "Welcome, Bruce Pelayo!" in resp.body
        resp = resp.click("Document", index=0)
        form = resp.forms['deform']
        form["title"] = "Bruce one"
        form.submit("save").maybe_follow()
        resp = app.get('/')
        resp = resp.click("Document", index=0)
        form = resp.forms['deform']
        form["title"] = "Bruce two"
        form.submit("save").maybe_follow()

        # Now let's delete Bruce::
        resp.click('Logout').maybe_follow()
        resp = self._login(app, 'admin', 'secret').maybe_follow()
        resp = resp.click("User Management")
        form = resp.forms['form-principal-search']
        form["query"] = "Bruce"
        resp = form.submit("search").maybe_follow()
        resp = resp.click("Bruce Pelayo")
        resp = resp.forms['deform'].submit('delete').maybe_follow()
        assert "Delete <em>Bruce Pelayo</em>" in resp.body
        assert "Are you sure you want to delete <span>User</span> " \
               "<em>Bruce Pelayo</em>?" in resp.body
        resp = resp.forms['form-delete-user'].submit('delete').maybe_follow()
        assert "User Bruce Pelayo was deleted." in resp.body
        form = resp.forms['form-principal-search']
        form["query"] = "Bruce"
        resp = form.submit("search").maybe_follow()
        assert "No users or groups were found." in resp.body



# -*- coding: utf-8 -*-
import datetime
import time

from mock import patch
from mock import MagicMock
import pytest

from kotti.resources import File
from kotti.resources import Image
from kotti.testing import asset
from kotti.testing import Dummy
from kotti.views.cache import set_max_age


def parse_expires(date_string):
    return datetime.datetime(*(
        time.strptime(date_string, "%a, %d %b %Y %H:%M:%S GMT")[0:6]))


def delta(date_string):
    now = datetime.datetime.utcnow()
    return parse_expires(date_string) - now


@pytest.fixture
def cachetest_content(root, filedepot):
    image = asset('sendeschluss.jpg')
    root['textfile'] = File("file contents", u"mytext.txt", u"text/plain")
    root['image'] = Image(image.read(), u"sendeschluss.jpg", u"image/jpeg")


class TestSetMaxAge:
    def test_preserve_existing_headers(self):
        response = Dummy(headers={
            "cache-control": "max-age=17,s-max-age=42,foo,bar=42"})
        delta = datetime.timedelta(days=1)
        set_max_age(response, delta)

        assert "cache-control" in response.headers
        assert response.headers["cache-control"] == (
            "bar=42,foo,max-age=86400,s-max-age=42")


class TestSetCacheHeaders:
    def test_caching_policy_chooser(self):
        from kotti.views.cache import set_cache_headers

        with patch('kotti.views.cache.caching_policy_chooser') as chooser:
            chooser.return_value = 'Random policy'

            event = MagicMock()
            event.response.headers.get.return_value = None

            with pytest.raises(KeyError):
                set_cache_headers(event)

        chooser.assert_called_with(
            event.request.context, event.request, event.response)

    def test_caching_policy_chooser_raises(self):
        from kotti.views.cache import set_cache_headers

        def raiser(*args, **kw):
            raise Exception()

        with patch('kotti.views.cache.caching_policy_chooser') as chooser:
            chooser.side_effect = raiser

            event = MagicMock()
            event.response.headers.get.return_value = None

            with patch('kotti.views.cache.logger'):
                set_cache_headers(event)

        chooser.assert_called_with(
            event.request.context, event.request, event.response)

    def test_header_set_before(self):
        from kotti.views.cache import CACHE_POLICY_HEADER
        from kotti.views.cache import set_cache_headers

        event = MagicMock()
        event.response.headers = {CACHE_POLICY_HEADER: 'Random policy'}

        with patch('kotti.views.cache.caching_policy_chooser') as chooser:
            with pytest.raises(KeyError):
                set_cache_headers(event)

        assert chooser.call_count == 0

    def test_request_has_no_context(self):
        from kotti.views.cache import set_cache_headers

        with patch('kotti.views.cache.caching_policy_chooser') as chooser:
            event = MagicMock()
            event.request = {}
            set_cache_headers(event)

            assert chooser.call_count == 0


class TestBrowser:

    def test_cache_unauth(self, webtest, cachetest_content):

        # html
        resp = webtest.app.get('/')
        assert resp.headers.get('X-Caching-Policy') == 'Cache HTML'
        assert resp.headers.get('Cache-Control') == 'max-age=0,s-maxage=3600'
        d = delta(resp.headers.get('Expires'))
        assert (d.days, d.seconds) < (0, 0)

        # resources
        resp = webtest.app.get('/static-kotti/base.css')
        assert resp.headers.get('X-Caching-Policy') == 'Cache Resource'
        assert resp.headers.get('Cache-Control') == 'max-age=2764800,public'
        d = delta(resp.headers.get('Expires'))
        assert (d.days, d.seconds) > (30, 0)
        assert 'Last-Modified' in resp.headers

        # post
        resp = webtest.app.post('/', '')
        assert 'X-Caching-Policy' not in resp.headers

        # 404
        resp = webtest.app.get('/this-isnt-here', status=404)
        assert 'X-Caching-Policy' not in resp.headers

        # media content
        resp = webtest.app.get('/textfile/inline-view')
        assert resp.headers.get('X-Caching-Policy') == 'Cache Media Content'
        assert resp.headers.get('Cache-Control') == 'max-age=14400,public'
        d = delta(resp.headers.get('Expires'))
        assert (d.days, d.seconds) > (0, 14000)
        resp = webtest.app.get('/image/inline-view')
        assert resp.headers.get('X-Caching-Policy') == 'Cache Media Content'
        assert resp.headers.get('Cache-Control') == 'max-age=14400,public'
        d = delta(resp.headers.get('Expires'))
        assert (d.days, d.seconds) > (0, 14000)

    @pytest.mark.user('admin')
    def test_cache_auth(self, webtest, cachetest_content):

        # html
        resp = webtest.app.get('/')
        assert resp.headers.get('X-Caching-Policy') == 'No Cache'

        # resources
        resp = webtest.app.get('/static-kotti/base.css')
        assert resp.headers.get('X-Caching-Policy') == 'Cache Resource'

        # 404
        resp = webtest.app.get('/this-isnt-here', status=404)
        assert 'X-Caching-Policy' not in resp.headers

        # media content
        resp = webtest.app.get('/textfile/inline-view')
        resp.headers.get('X-Caching-Policy') == 'No Cache'
        assert resp.headers.get('Cache-Control') == 'max-age=0,public'
        d = delta(resp.headers.get('Expires'))
        assert (d.days, d.seconds) <= (0, 0)
        resp = webtest.app.get('/image/inline-view')
        resp.headers.get('X-Caching-Policy') == 'No Cache'
        assert resp.headers.get('Cache-Control') == 'max-age=0,public'
        d = delta(resp.headers.get('Expires'))
        assert (d.days, d.seconds) <= (0, 0)



# -*- coding: utf-8 -*-

import warnings

from pyramid.httpexceptions import HTTPNotFound
from pyramid.view import notfound_view_config
from pyramid.view import render_view_to_response
from pyramid.view import view_config

from kotti.interfaces import IContent
from kotti.views.util import search_content
from kotti.views.util import search_content_for_tags


@view_config(context=IContent)
def view_content_default(context, request):
    """This view is always registered as the default view for any Content.

    Its job is to delegate to a view of which the name may be defined
    per instance.  If a instance level view is not defined for
    'context' (in 'context.defaultview'), we will fall back to a view
    with the name 'view'.
    """

    view_name = context.default_view or 'view'
    response = render_view_to_response(context, request, name=view_name)
    if response is None:  # pragma: no cover
        warnings.warn(
            u'Failed to look up default view called {0!r} for {1!r}.'.format(
                view_name, context))
        raise HTTPNotFound()
    return response


# noinspection PyUnusedLocal
def view_node(context, request):  # pragma: no cover
    return {}  # BBB


# noinspection PyUnusedLocal
@view_config(name='search-results', permission='view',
             renderer='kotti:templates/view/search-results.pt')
def search_results(context, request):
    results = []
    if u'search-term' in request.POST:
        search_term = request.POST[u'search-term']
        results = search_content(search_term, request)
    return {'results': results}


# noinspection PyUnusedLocal
@view_config(name='search-tag', permission='view',
             renderer='kotti:templates/view/search-results.pt')
def search_results_for_tag(context, request):
    results = []
    if u'tag' in request.GET:
        # Single tag searching only, is allowed in default Kotti. Add-ons can
        # utilize search_content_for_tags(tags) to enable multiple tags
        # searching, but here it is called with a single tag.
        tags = [request.GET[u'tag'].strip()]
        results = search_content_for_tags(tags, request)
    return {'results': results}


# noinspection PyUnusedLocal
@view_config(name='search', permission='view',
             renderer='kotti:templates/view/search.pt')
@view_config(name='folder_view', context=IContent, permission='view',
             renderer='kotti:templates/view/folder.pt')
@view_config(name='view', context=IContent, permission='view',
             renderer='kotti:templates/view/document.pt')
def view(context, request):
    return {}


@notfound_view_config(renderer='kotti:templates/http-errors/notfound.pt')
def notfound_view(context, request):
    request.response = context
    return {}


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.scan(__name__)



# -*- coding: utf-8 -*-
"""
Content edit views
"""

import random
from StringIO import StringIO
import colander
from colander import SchemaNode
from colander import null
from deform import FileData
from deform.widget import FileUploadWidget
from deform.widget import RichTextWidget
from deform.widget import TextAreaWidget

from kotti.resources import Document
from kotti.resources import File
from kotti.resources import Node
# noinspection PyProtectedMember
from kotti.util import _
# noinspection PyProtectedMember
from kotti.util import _to_fieldstorage
from kotti.views.form import get_appstruct
from kotti.views.form import AddFormView
from kotti.views.form import EditFormView
from kotti.views.form import FileUploadTempStore
from kotti.views.form import ObjectType
from kotti.views.form import deferred_tag_it_widget
from kotti.views.form import validate_file_size_limit


class ContentSchema(colander.MappingSchema):
    title = colander.SchemaNode(
        colander.String(),
        title=_(u'Title'),
        validator=colander.Length(
            max=Node.title.property.columns[0].type.length),
        )
    description = colander.SchemaNode(
        colander.String(),
        title=_('Description'),
        widget=TextAreaWidget(cols=40, rows=5),
        missing=u"",
        )
    tags = colander.SchemaNode(
        ObjectType(),
        title=_('Tags'),
        widget=deferred_tag_it_widget,
        missing=[],
        )


class DocumentSchema(ContentSchema):
    body = colander.SchemaNode(
        colander.String(),
        title=_(u'Body'),
        widget=RichTextWidget(
            # theme='advanced', width=790, height=500
            height=500,
        ),
        missing=u"",
        )


# noinspection PyPep8Naming
def FileSchema(tmpstore, title_missing=None):
    class FileSchema(ContentSchema):
        file = SchemaNode(
            FileData(),
            title=_(u'File'),
            widget=FileUploadWidget(tmpstore),
            validator=validate_file_size_limit,
            )

    # noinspection PyUnusedLocal
    def set_title_missing(node, kw):
        if title_missing is not None:
            node['title'].missing = title_missing

    return FileSchema(after_bind=set_title_missing)


class DocumentEditForm(EditFormView):
    schema_factory = DocumentSchema


class DocumentAddForm(AddFormView):
    schema_factory = DocumentSchema
    add = Document
    item_type = _(u"Document")


class FileEditForm(EditFormView):
    def before(self, form):
        form.appstruct = get_appstruct(self.context, self.schema)
        if self.context.data is not None:
            form.appstruct.update({'file': {
                'fp': StringIO(self.context.data.file.read()),
                'filename': self.context.name,
                'mimetype': self.context.mimetype,
                'uid': str(random.randint(1000000000, 9999999999)),
            }})

    def schema_factory(self):
        tmpstore = FileUploadTempStore(self.request)
        return FileSchema(tmpstore)

    def edit(self, **appstruct):
        title = appstruct['title']
        self.context.title = title
        self.context.description = appstruct['description']
        self.context.tags = appstruct['tags']
        if appstruct['file']:
            self.context.data = _to_fieldstorage(**appstruct['file'])


class FileAddForm(AddFormView):
    item_type = _(u"File")
    item_class = File  # specific to this class

    def schema_factory(self):
        tmpstore = FileUploadTempStore(self.request)
        return FileSchema(tmpstore, title_missing=null)

    def save_success(self, appstruct):
        if not appstruct['title']:
            appstruct['title'] = appstruct['file']['filename']
        return super(FileAddForm, self).save_success(appstruct)

    def add(self, **appstruct):
        filename = appstruct['file']['filename']
        item = self.item_class(
            title=appstruct['title'] or filename,
            description=appstruct['description'],
            tags=appstruct['tags'],
            data=_to_fieldstorage(**appstruct['file']),
        )
        return item


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.add_view(
        DocumentEditForm,
        context=Document,
        name='edit',
        permission='edit',
        renderer='kotti:templates/edit/node.pt',
        )

    config.add_view(
        DocumentAddForm,
        name=Document.type_info.add_view,
        permission=Document.type_info.add_permission,
        renderer='kotti:templates/edit/node.pt',
        )

    config.add_view(
        FileEditForm,
        context=File,
        name='edit',
        permission='edit',
        renderer='kotti:templates/edit/node.pt',
        )

    config.add_view(
        FileAddForm,
        name=File.type_info.add_view,
        permission=File.type_info.add_permission,
        renderer='kotti:templates/edit/node.pt',
        )

# DEPRECATED

# noinspection PyPep8
from zope.deprecation import deprecated
# noinspection PyPep8
from kotti_image.views.edit import ImageAddForm
# noinspection PyPep8
from kotti_image.views.edit import ImageEditForm
__ = ImageAddForm, ImageEditForm   # pyflakes

deprecated(('ImageAddForm', 'ImageEditForm'),
           'Image was outfactored to the kotti_image package.  '
           'Please import from there.')



# -*- coding: utf-8 -*-

"""
Inheritance Diagram
-------------------

.. inheritance-diagram:: kotti.util
"""

import cgi
import re
try:
    # PY2
    from urllib import unquote
    from urlparse import urlparse
    from urlparse import urlunparse
except ImportError:
    from urllib.parse import unquote
    from urllib.parse import urlparse
    from urllib.parse import urlunparse


from docopt import docopt
from pyramid.i18n import get_localizer
from pyramid.i18n import get_locale_name
from pyramid.i18n import make_localizer
from pyramid.i18n import TranslationStringFactory
from pyramid.interfaces import ITranslationDirectories
from pyramid.location import inside
from pyramid.paster import bootstrap
from pyramid.paster import setup_logging
from pyramid.renderers import render
from pyramid.threadlocal import get_current_registry
from pyramid.threadlocal import get_current_request
from pyramid.url import resource_url
from pyramid.view import render_view_to_response
from repoze.lru import LRUCache
from zope.deprecation import deprecated

from kotti import DBSession

_ = TranslationStringFactory('Kotti')


def get_localizer_for_locale_name(locale_name):
    registry = get_current_registry()
    tdirs = registry.queryUtility(ITranslationDirectories, default=[])
    return make_localizer(locale_name, tdirs)


def translate(*args, **kwargs):
    request = get_current_request()
    if request is None:
        localizer = get_localizer_for_locale_name('en')
    else:
        localizer = get_localizer(request)
    return localizer.translate(*args, **kwargs)


def get_paste_items(context, request):
    from kotti.resources import Node

    items = []
    info = request.session.get('kotti.paste')
    if info:
        ids, action = info
        for id in ids:
            item = DBSession.query(Node).get(id)
            if item is None or not item.type_info.addable(context, request):
                continue
            if action == 'cut' and inside(context, item):
                continue
            if context == item:
                continue
            items.append(item)
    return items


def render_view(context, request, name='', secure=True):
    from kotti.security import authz_context

    with authz_context(context, request):
        response = render_view_to_response(context, request, name, secure)
    if response is not None:
        return response.ubody


class TemplateStructure(object):
    def __init__(self, html):
        self.html = html

    def __html__(self):
        return self.html
    __unicode__ = __html__

    def __getattr__(self, key):
        return getattr(self.html, key)


class LinkBase(object):
    def __call__(self, context, request):
        return TemplateStructure(
            render(
                self.template,
                dict(link=self, context=context, request=request),
                request,
                )
            )

    def selected(self, context, request):
        """ Returns True if the Link's url, based on its name,
        matches the request url

        If the link name is '', it will be selected for all urls ending in '/'
        """
        parsed = urlparse(unquote(request.url))

        # insert view markers @@ in last component of the path
        path = parsed.path.split('/')
        if '@@' not in path[-1]:
            path[-1] = '@@' + path[-1]
        path = '/'.join(path)
        url = urlunparse((parsed[0], parsed[1], path, '', '', ''))

        return url == self.url(context, request)

    def permitted(self, context, request):
        from kotti.security import view_permitted
        return view_permitted(context, request, self.name)

    def visible(self, context, request):
        permitted = self.permitted(context, request)
        if permitted:
            if self.predicate is not None:
                return self.predicate(context, request)
            else:
                return True
        return False

    @property
    def path(self):  # BBB
        return self.name
    path = deprecated(
        path,
        "The 'path' attribute has been deprecated as of Kotti 1.0.0.  Please "
        "use 'name' instead.",
        )


class LinkRenderer(LinkBase):
    """A menu link that renders a view to render the link.
    """
    def __init__(self, name, predicate=None):
        self.name = name
        self.predicate = predicate

    def __call__(self, context, request):
        return TemplateStructure(render_view(context, request, name=self.name))

    # noinspection PyMethodOverriding
    @staticmethod
    def selected(context, request):
        return False


class LinkParent(LinkBase):
    """A menu link that renders sublinks in a dropdown.
    """
    template = 'kotti:templates/edit/el-parent.pt'

    def __init__(self, title, children):
        self.title = title
        self.children = children

    def visible(self, context, request):
        return any(ch.visible(context, request) for ch in self.children)

    def selected(self, context, request):
        return any(ch.selected(context, request) for ch in self.children)

    def get_visible_children(self, context, request):
        return [ch for ch in self.children if ch.visible(context, request)]


class Link(LinkBase):
    template = 'kotti:templates/edit/el-link.pt'

    def __init__(self, name, title=None, predicate=None, target=None):
        self.name = name
        if title is None:
            title = name.replace('-', ' ').replace('_', ' ').title()
        self.title = title
        self.predicate = predicate
        self.target = target

    def url(self, context, request):
        return resource_url(context, request) + '@@' + self.name

    def __eq__(self, other):
        return isinstance(other, Link) and repr(self) == repr(other)

    def __repr__(self):
        return u'Link({0}, {1})'.format(self.name, self.title)


class ActionButton(Link):
    def __init__(self, path, title=None, no_children=False,
                 css_class=u"btn btn-default"):
        super(ActionButton, self).__init__(path, title)
        self.no_children = no_children
        self.css_class = css_class


class DontCache(Exception):
    pass

_CACHE_ATTR = 'kotti_cache'


def request_container():
    request = get_current_request()
    if request is None:
        return None
    cache = getattr(request, _CACHE_ATTR, None)
    if cache is None:
        cache = {}
        setattr(request, _CACHE_ATTR, cache)
    return cache


def cache(compute_key, container_factory):
    marker = object()

    def decorator(func):
        def replacement(*args, **kwargs):
            cache = container_factory()
            if cache is None:
                return func(*args, **kwargs)
            try:
                key = compute_key(*args, **kwargs)
            except DontCache:
                return func(*args, **kwargs)
            key = u'{0}.{1}:{2}'.format(func.__module__, func.__name__, key)
            cached_value = cache.get(key, marker)
            if cached_value is marker:
                cached_value = cache[key] = func(*args, **kwargs)
            else:
                pass
            return cached_value
        replacement.__doc__ = func.__doc__
        return replacement
    return decorator


def request_cache(compute_key):
    return cache(compute_key, request_container)


class LRUCacheSetItem(LRUCache):
    __setitem__ = LRUCache.put

_lru_cache = LRUCacheSetItem(1000)


def lru_cache(compute_key):
    return cache(compute_key, lambda: _lru_cache)


def clear_cache():  # only useful for tests really
    request = get_current_request()
    if request is not None:
        setattr(request, _CACHE_ATTR, None)
    _lru_cache.clear()


def extract_from_settings(prefix, settings=None):
    """
      >>> settings = {
      ...     'kotti_twitter.foo_bar': '1', 'kotti.spam_eggs': '2'}
      >>> print(extract_from_settings('kotti_twitter.', settings))
      {'foo_bar': '1'}
    """
    from kotti import get_settings
    settings = settings if settings is not None else get_settings()
    extracted = {}
    for key, value in settings.items():
        if key.startswith(prefix):
            extracted[key[len(prefix):]] = value
    return extracted


def disambiguate_name(name):
    parts = name.split(u'-')
    if len(parts) > 1:
        try:
            index = int(parts[-1])
        except ValueError:
            parts.append(u'1')
        else:
            parts[-1] = unicode(index + 1)
    else:
        parts.append(u'1')
    return u'-'.join(parts)


def title_to_name(title, blacklist=(), max_length=None):
    """ If max_length is None, fallback to the ``name`` column
        size (:class:`kotti.resources.Node`)
    """
    if max_length is None:
        from kotti.resources import Node
        # See #428, #427 and #31
        max_length = Node.name.property.columns[0].type.length - 10
    request = get_current_request()
    if request is not None:
        locale_name = get_locale_name(request)
    else:
        locale_name = 'en'
    from kotti import get_settings
    urlnormalizer = get_settings()['kotti.url_normalizer'][0]
    name = unicode(urlnormalizer(title, locale_name, max_length=max_length))
    if name not in blacklist:
        return name
    name += u'-1'
    while name in blacklist:
        name = disambiguate_name(name)
    return name


def camel_case_to_name(text):
    """
      >>> camel_case_to_name('FooBar')
      'foo_bar'
      >>> camel_case_to_name('TXTFile')
      'txt_file'
      >>> camel_case_to_name ('MyTXTFile')
      'my_txt_file'
      >>> camel_case_to_name('froBOZ')
      'fro_boz'
      >>> camel_case_to_name('f')
      'f'
    """
    return re.sub(
        r'((?<=[a-z])[A-Z]|(?<!\A)[A-Z](?=[a-z]))', r'_\1', text).lower()


def command(func, doc):
    args = docopt(doc)
    # establish config file uri
    config_uri = args['<config_uri>']
    pyramid_env = bootstrap(config_uri)
    # Setup logging to allow log output from command methods
    setup_logging(config_uri)
    try:
        func(args)
    finally:
        pyramid_env['closer']()
    return 0


ViewLink = Link
deprecated(
    'ViewLink',
    "kotti.util.ViewLink has been renamed to Link as of Kotti 1.0.0."
    )


def _to_fieldstorage(fp, filename, mimetype, size, **_kwds):
    """ Build a :class:`cgi.FieldStorage` instance.

    Deform's :class:`FileUploadWidget` returns a dict, but
    :class:`depot.fields.sqlalchemy.UploadedFileField` likes
    :class:`cgi.FieldStorage` objects
    """
    f = cgi.FieldStorage()
    f.file = fp
    f.filename = filename
    f.type = mimetype
    f.length = size
    return f



# -*- coding: utf-8 -*-
"""Update Node.path column

Revision ID: 559ce6eb0949
Revises: 1063d7178fa
Create Date: 2014-12-10 13:20:29.374951

"""

# revision identifiers, used by Alembic.
revision = '559ce6eb0949'
down_revision = '1063d7178fa'


def upgrade():

    from kotti.resources import DBSession

    from alembic.context import get_bind

    conn = get_bind()

    if conn.engine.dialect.name == 'mysql':
        update = "update nodes set path = concat(path, '/') where path not \
like '%/'"
    else:
        update = "update nodes set path = path || '/' where path not like '%/'"
    DBSession.execute(update)


def downgrade():
    from kotti import DBSession
    from kotti.resources import Node

    for node in DBSession.query(Node).with_polymorphic([Node]):
        # remove trailing '/' from all nodes but root
        if node.path != u'/':
            node.path = node.path[:-1]



# -*- coding: utf-8 -*-
""" Kotti HTTP Exception browser tests """

import pytest


def test_404_anon(webtest, root):

    resp = webtest.app.get('/non-existent', status=404)
    assert 'Not Found' in resp.text


@pytest.mark.user('admin')
def test_404_anon(webtest, root):

    resp = webtest.app.get('/non-existent', status=404)
    assert 'Not Found' in resp.text


def test_404_api_root(db_session, dummy_request):
    from kotti.resources import get_root
    from kotti.views.util import TemplateAPI
    from pyramid.httpexceptions import HTTPNotFound

    api = TemplateAPI(HTTPNotFound(), dummy_request)

    assert api.root == get_root()
